r"""Solve the `Cahn-Hilliard equation to generate data.

Solve the Cahn-Hilliard equation,
<https://en.wikipedia.org/wiki/Cahn-Hilliard_equation>`__, for
multiple samples in arbitrary dimensions. The concentration varies
from -1 to 1. The equation is given by

.. math::

   \dot{\phi} = \nabla^2 \left( \phi^3 -
                                \phi \right) - \gamma \nabla^4 \phi

The discretiztion scheme used here is from `Chang and Rutenberg
<http://dx.doi.org/10.1103/PhysRevE.72.055701>`__. The scheme is a
semi-implicit discretization in time and is given by

.. math::

   \phi_{t+\Delta t}
   + \left(1 - a_1\right) \Delta t \nabla^2 \phi_{t+\Delta t}
   + \left(1 - a_2\right) \Delta t \gamma \nabla^4 \phi_{t+\Delta t}
   = \phi_t
   - \Delta t \nabla^2 \left(a_1 \phi_t + a_2
                             \gamma \nabla^2 \phi_t - \phi_t^3 \right)

where :math:`a_1=3` and :math:`a_2=0`.

>>> import dask.array as da
>>> from toolz.curried import pipe

>>> solve_ = np_solve(gamma=1., delta_t=1.)

>>> def tester(shape, min_, max_, steps):
...     return pipe(
...         0.01 * (2 * da.random.random(shape, chunks=shape) - 1),
...         map_blocks(iterate_times(solve_, steps)),
...         lambda x: da.max(x) > max_ and da.min(x) < min_
...     )

1D

>>> da.random.seed(101)
>>> assert tester((2, 100), -0.9, 0.9, 100)

2D

>>> da.random.seed(101)
>>> assert tester((2, 101, 101), -0.9, 0.9, 100)

3D

>>> da.random.seed(101)
>>> assert tester((2, 101, 101, 101), -5e-4, 5e-4, 10)

"""

import numpy as np
from toolz.curried import memoize
from ..func import curry, map_blocks, ifftn, fftn, iterate_times


def _k_space(size):
    size1 = lambda: (size // 2) if (size % 2 == 0) else (size - 1) // 2
    size2 = lambda: size1() if (size % 2 == 0) else size1() + 1
    return np.concatenate(
        (np.arange(size)[: size2()], (np.arange(size) - size1())[: size1()])
    )


@memoize
def _calc_ksq_(shape):
    indices = lambda: np.indices(shape)
    return np.sum(_k_space(shape[0])[indices()] ** 2, axis=0)[None]


def _calc_ksq(x_data, delta_x):
    return _calc_ksq_(x_data.shape[1:]) * (2 * np.pi / (delta_x * x_data.shape[1])) ** 2


def _axes(x_data):
    return np.arange(len(x_data.shape) - 1) + 1


def _explicit(gamma, ksq, param_a1=3.0, param_a2=0.0):
    return param_a1 - gamma * param_a2 * ksq


def _f_response(x_data, delta_t, gamma, ksq):
    fx_data = lambda: fftn(x_data, axes=_axes(x_data))
    fx3_data = lambda: fftn(x_data ** 3, axes=_axes(x_data))
    implicit = lambda: (1 - gamma * ksq) - _explicit(gamma, ksq)
    delta_t_ksq = lambda: delta_t * ksq
    numerator = (
        lambda: fx_data() * (1 + delta_t_ksq() * _explicit(gamma, ksq))
        - delta_t_ksq() * fx3_data()
    )
    return numerator() / (1 - delta_t_ksq() * implicit())


@curry
def np_solve(x_data, delta_x=0.25, delta_t=0.001, gamma=1.0):
    """Solve the Cahn-Hilliard equation for one step.

    Advance multiple microstuctures in time with the Cahn-Hilliard
    equation.

    Args:
      x_data: the initial microstucture
      delta_x: the grid spacing
      delta_t: the time step size
      gamma: Cahn-Hilliard parameter

    Returns:
      an updated microsturcture
n

    Raises:
      RuntimeError if domain is not square

    """
    return ifftn(
        _f_response(_check(x_data), delta_t, gamma, _calc_ksq(x_data, delta_x)),
        axes=_axes(x_data),
    ).real


def _check(x_data):
    """Ensure that domain is square.

    Args:
      x_data: the initial microstuctures

    Returns:
      the initial microstructures

    Raises:
      RuntimeError if microstructures are not square

    >>> _check(np.array([[[1, 2, 3], [4, 5, 6]]]))
    Traceback (most recent call last):
    ...
    RuntimeError: X must represent a square domain

    """
    if not np.all(np.array(x_data.shape[1:]) == x_data.shape[1]):
        raise RuntimeError("X must represent a square domain")

    return x_data


def solve_cahn_hilliard(x_data, n_steps=1, delta_x=0.25, delta_t=0.001, gamma=1.0):
    r"""Solve the Cahn-Hilliard equation.

    Solve the `Cahn-Hilliard equation
    <https://en.wikipedia.org/wiki/Cahn-Hilliard_equation>`_ for
    multiple samples in arbitrary dimensions. The concentration varies
    from -1 to 1. The equation is given by

    .. math::

       \dot{\phi} = \nabla^2 \left( \phi^3 - \phi \right) - \gamma \nabla^4 \phi

    The discretiztion scheme used here is from `Chang and Rutenberg
    <http://dx.doi.org/10.1103/PhysRevE.72.055701>`_. The scheme is a
    semi-implicit discretization in time and is given by

    .. math::

       \phi_{t+\Delta t}
       + \left(1 - a_1\right) \Delta t \nabla^2 \phi_{t+\Delta t}
       + \left(1 - a_2\right) \Delta t \gamma \nabla^4 \phi_{t+\Delta t}
       = \phi_t
       - \Delta t \nabla^2 \left(a_1 \phi_t + a_2
          \gamma \nabla^2 \phi_t - \phi_t^3 \right)

    where :math:`a_1=3` and :math:`a_2=0`.

    Args:
      x_data: dask array chunked along the sample axis ``(n_sample, n_x, n_y)``
      n_steps: number of time steps used
      delta_x: the grid spacing, :math:`\Delta x`
      delta_t: the time step size, :math:`\Delta t`
      gamma: Cahn-Hilliard parameter, :math:`\gamma`

    >>> import dask.array as da
    >>> da.random.seed(99)
    >>> x_data = 2 * da.random.random((1, 100, 100), chunks=(1, 100, 100)) - 1
    >>> y_data = solve_cahn_hilliard(x_data)
    >>> y_data.chunks
    ((1,), (100,), (100,))

    >>> y_data = solve_cahn_hilliard(x_data, n_steps=10000)  #doctest: +SKIP
    >>> from pymks import plot_microstructures
    >>> fig = plot_microstructures(x_data[0], y_data[0])
    >>> fig.show()  #doctest: +SKIP

    .. image:: cahn-hilliard.png
       :width: 400

    """
    return map_blocks(
        iterate_times(np_solve(delta_x=delta_x, delta_t=delta_t, gamma=gamma), n_steps),
        x_data,
    )



"""Test plotting functionality
"""

import matplotlib
import numpy as np
from pymks.fmks.plot import plot_microstructures

matplotlib.use("Agg")


def test_plot():
    """Ensure that the plot_micostructures function is run during the
    tests.

    """
    plot_microstructures(np.arange(4).reshape(2, 2), titles=["test"])
    plot_microstructures(np.arange(4).reshape(2, 2), titles="test")
    plot_microstructures(np.arange(4).reshape(2, 2))



"""Test the primitive basis.
"""

import numpy as np
from toolz import pipe
from pymks.fmks.bases.primitive import discretize


def test_local():
    """Local state example.

    An examples where the local states fall between the picks of local
    states. The first specifies the local state space domain between
    `[0, 1]`.
    """
    np.random.seed(4)

    def _compare(x_raw, n_state):
        basis = discretize(n_state=n_state)
        return pipe(
            np.linspace(0, 1, n_state),
            lambda h: np.sum(basis(x_raw) * h[None, None, None, :], axis=-1),
            lambda x: np.allclose(x_raw, x),
        )

    assert _compare(np.random.random((2, 5, 3, 2)), 10)


def test_local_min_max():
    """Local state example with varying min and max.

    An example where the local state space domain is between `[-1,
    1]`.
    """
    basis = discretize(n_state=3, min_=-1)
    assert np.allclose(
        basis(np.array([-1, 0, 1, 0.5])),
        [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0.5, 0.5]],
    )


def test_error():
    """Test data outside the bounds

    If the local state values in the microstructure are outside of the
    domain they are remapped inside of the domain
    """
    basis = discretize(n_state=2)
    assert np.allclose(basis(np.array([-1, 1])), [[1, 0], [0, 1]])


def test_many_local_states():
    """Test with many local states
    """

    def _x_data():
        np.random.seed(3)
        return np.random.random((2, 5, 3))

    def _test_data(n_state):
        basis = discretize(n_state=n_state)
        states = lambda: np.linspace(0, 1, n_state)[None, None, None, :]
        return np.sum(basis(_x_data()) * states(), axis=-1)

    assert np.allclose(_x_data(), _test_data(10))



from .gsh_functions import hex_eval
from .gsh_functions import cub_eval
from .gsh_functions import tri_eval
from .gsh_functions import hex_basis_info
from .gsh_functions import cub_basis_info
from .gsh_functions import tri_basis_info
from .imag_ffts import _ImagFFTBasis
import numpy as np
from ..fmks.func import deprecate


class GSHBasis(_ImagFFTBasis):

    r"""
    Discretize a continuous field made up three Euler angles (in radians) used
    to represent crystal orientation into continuous local states using the
    Generalized Spherical Harmonic (GSH) basis. This basis uses the following
    equation to discretize the orientation field.

    .. math::

       \frac{1}{\Delta x} \int_s m(g, x) dx =
       \sum_{l, m, n} m[l, \tilde{m}, n, s] T_l^{\tilde{m}n}(g)

    where the :math:`T_l^{\tilde{m}n}` are GSH basis functions and the
    local state space :math:`H` is mapped into the orthogonal, periodic
    domain of the GSH functions

    The mapping of :math:`H` into some desired periodic domain is done
    automatically in PyMKS by using the `domain` key work argument to
    select the desired crystal symmetry.

    >>> X = np.array([[0.1, 0.2, 0.3],
    ...               [6.5, 2.3, 3.4]])
    >>> gsh_basis = GSHBasis(n_states = [3], domain='hexagonal')
    >>> def test_gsh(x):
    ...     phi = x[:, 1]
    ...     t915 = np.cos(phi)
    ...     return 0.15e2 / 0.2e1 * t915 ** 2 - 0.5e1 / 0.2e1

    >>> assert(np.allclose(np.squeeze(gsh_basis.discretize(X)), test_gsh(X)))

    If you select an invalid crystal symmetry PyMKS will give an error

    >>> gsh_basis = GSHBasis(n_states=[3], domain='squishy')
    ... # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    RuntimeError: invalid crystal symmetry

    >>> gsh_basis = GSHBasis(n_states=[3], domain='hex') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    RuntimeError: invalid crystal symmetry
    """
    @deprecate
    def __init__(self, n_states=15, domain=None):
        """
        Instantiate a `Basis`

        Args:
            n_states (int, array): An array of local states to be used. states
                requested. If an integer is provided, all local states up
                to that number will be used.
            domain (list, optional): indicate the desired crystal symmetry for
                the GSH. Valid choices for symmetry are "hexagonal", "cubic" or
                "triclinic" if no symmetry is desired (not specifying any
                symmetry has the same effect)
        """
        if isinstance(n_states, int):
            n_states = np.arange(n_states)
        if domain is None or domain == 'triclinic':
            self.domain = 'triclinic'
        elif domain == 'hexagonal':
            self.domain = 'hexagonal'
        elif domain == 'cubic':
            self.domain = 'cubic'
        else:
            raise RuntimeError("invalid crystal symmetry")
        full_indx = self._gsh_basis_info()
        super(GSHBasis, self).__init__(n_states=n_states, domain=self.domain)
        self.basis_indices = full_indx[self.n_states, :]

    def check(self, X):
        """Warns the user if Euler angles apear to be defined in degrees
        instead of radians"""
        if (np.min(X) < -90.) or (np.max(X) > 90.):
            Warning("X may be defined in degrees instead of radians")
        if X.shape[-1] != 3:
            raise RuntimeError('X must have 3 angles (in radians) in the ' +
                               'last dimention')

    def _check_shape(self, X_shape, y_shape):
        """
        Checks the shape of the microstructure and response data to
        ensure that they are correct.

        Firstly, the response data "y" must have a dimension to index the
        microstructure instantiation and at least one dimension to index the
        local microstructural information.

        Second, the shape of X and y must me equal except for the last
        dimension of X.

        Finally, the length of the final dimension of X must be 3.
        This is because we assume that Bunge Euler angles are assigned for
        each location in the microstructure
        """
        if not len(y_shape) > 1:
            raise RuntimeError("The shape of y is incorrect.")
        if X_shape[-1] != 3:
            raise RuntimeError("X must have 3 continuous local states " +
                               "(euler angles in radians) in the last axis.")
        if y_shape != X_shape[:-1]:
            raise RuntimeError("The X and y must have the same number of " +
                               "samples and microstructure shape.")

    def _pred_shape(self, X):
        """
        Function to describe the expected output shape of a given
        microstructure X.
        """
        _shape = X.shape[:-1]  # X has Euler angles, while output is scalar
        if len(_shape) < 2:
            _shape = (X.shape[0],) + (X.shape[1] / 3,)
        return _shape

    def discretize(self, X):
        """
        Discretize `X`.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ..., 3)`
                shaped array where `n_samples` is the number of samples,
                `n_x` is the spatial discretization and the last dimension
                contains the Bunge Euler angles in radians.
        Returns:
            Float valued field of of Generalized Spherical Harmonics
            coefficients.

        >>> X = np.array([[0.1, 0.2, 0.3],
        ...               [6.5, 2.3, 3.4]])
        >>> gsh_basis = GSHBasis(n_states = [1])
        >>> def q(x):
        ...     phi1 = x[:, 0]
        ...     phi = x[:, 1]
        ...     phi2 = x[:, 2]
        ...     x_GSH = ((0.3e1 / 0.2e1) * (0.1e1 + np.cos(phi)) *
        ...              np.exp((-1*1j) * (phi1 + phi2)))
        ...     return x_GSH
        >>> assert(np.allclose(np.squeeze(gsh_basis.discretize(X)), q(X)))
        """
        self.check(X)
        self._select_axes(X[..., 0])
        return self._gsh_eval(X)

    def _reshape_feature(self, X, size):
        """
        Helper function used to check the shape of the microstructure,
        and change to appropriate shape.

        Args:
            X: The microstructure, an `(n_samples, n_x, ...)` shaped array
                where `n_samples` is the number of samples and `n_x` is thes
                patial discretization.
            size: the new size of the array

        Returns:
            microstructure with shape (n_samples, size)
        """
        _shape = (X.shape[0],) + size + (3,)
        return X.reshape(_shape)

    def _gsh_basis_info(self):
        """
        Returns a an array with the indices used in the GSH functions for a
        a specific crystal symmetry.
        """
        if self.domain == 'triclinic':
            return tri_basis_info()
        elif self.domain == 'hexagonal':
            return hex_basis_info()
        elif self.domain == 'cubic':
            return cub_basis_info()

    def _gsh_eval(self, X):
        """
        Discretize `X` based on crystal symmetry.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ..., 3)`
                shaped array where `n_samples` is the number of samples,
                `n_x` is the spatial discretization and the last dimension
                contains the Bunge Euler angles in radians.
        Returns:
            Float valued field of of Generalized Spherical Harmonics
            coefficients.
        """
        if self.domain == 'triclinic':
            return tri_eval(X, self.n_states)
        elif self.domain == 'hexagonal':
            return hex_eval(X, self.n_states)
        elif self.domain == 'cubic':
            return cub_eval(X, self.n_states)



from .real_ffts import _RealFFTBasis
import numpy as np


class LegendreBasis(_RealFFTBasis):

    r"""
    Discretize a continuous field into `deg` local states using a
    Legendre polynomial basis such that,

    .. math::

       \frac{1}{\Delta x} \int_s m(h, x) dx =
       \sum_0^{L-1} m[l, s] P_l(h)

    where the :math:`P_l` are Legendre polynomials and the local state space
    :math:`H` is mapped into the orthogonal domain of the Legendre polynomials

    .. math::

       -1 \le  H \le 1

    The mapping of :math:`H` into the domain is done automatically in PyMKS by
    using the `domain` key work argument.

    >>> n_states = 3
    >>> X = np.array([[0.25, 0.1],
    ...               [0.5, 0.25]])
    >>> def P(x):
    ...    x = 4 * x - 1
    ...    polys = np.array((np.ones_like(x), x, (3.*x**2 - 1.) / 2.))
    ...    tmp = (2. * np.arange(3)[:, None, None] + 1.) / 2. * polys
    ...    return np.rollaxis(tmp, 0, 3)
    >>> basis = LegendreBasis(n_states, [0., 0.5])
    >>> assert(np.allclose(basis.discretize(X), P(X)))

    If the microstructure local state values fall outside of the specified
    domain they will no longer be mapped into the orthogonal domain of the
    legendre polynomials.

    >>> n_states = 2
    >>> X = np.array([-1, 1])
    >>> leg_basis = LegendreBasis(n_states, domain=[0, 1])
    >>> leg_basis.discretize(X)
    Traceback (most recent call last):
    ...
    RuntimeError: X must be within the specified domain

    """

    def discretize(self, X):
        """
        Discretize `X`.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ...)`
                shaped array where `n_samples` is the number of samples and
                `n_x` is the spatial discretization.
        Returns:
            Float valued field of of Legendre polynomial coefficients.

        >>> X = np.array([[-1, 1],
        ...               [0, -1]])
        >>> leg_basis = LegendreBasis(3, [-1, 1])
        >>> def p(x):
        ...    polys = np.array((np.ones_like(x), x, (3.*x**2 - 1.) / 2.))
        ...    tmp = (2. * np.arange(3)[:, None, None] + 1.) / 2. * polys
        ...    return np.rollaxis(tmp, 0, 3)
        >>> assert(np.allclose(leg_basis.discretize(X), p(X)))

        """
        self.check(X)
        self._select_axes(X)
        leg = np.polynomial.legendre
        X_scaled = (2. * X - self.domain[0] - self.domain[1]) /\
                   (self.domain[1] - self.domain[0])
        norm = (2. * np.array(self.n_states) + 1) / 2.
        X_Legendre = (leg.legval(X_scaled, np.eye(len(self.n_states)) * norm))
        return np.rollaxis(X_Legendre, 0, len(X_Legendre.shape))



"""
The MKS homogenization module test cases

"""
import numpy as np
import dask.array as da
from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA
from sklearn.linear_model import LogisticRegression
from pymks.fmks.correlations import FlattenTransformer, TwoPointCorrelation
from pymks.fmks.data.cahn_hilliard import solve_cahn_hilliard
from pymks.fmks.bases.legendre import LegendreTransformer


def test_classification():
    """This test basically creates Legendre microstructures in both times:
    0 and t.  Then builds homogenization classification linkages to
    classify if newly generated microstructures are at time 0 or t
    """
    reducer = PCA(n_components=3)
    linker = LogisticRegression()
    homogenization_pipeline = Pipeline(
        steps=[
            ("discretize", LegendreTransformer(n_state=3, min_=-1.0, max_=1.0)),
            (
                "Correlations",
                TwoPointCorrelation(
                    periodic_boundary=True, cutoff=10, correlations=[(1, 1), (0, 1)]
                ),
            ),
            ("flatten", FlattenTransformer()),
            ("reducer", reducer),
            ("connector", linker),
        ]
    )
    da.random.seed(3)
    x0_phase = 2 * da.random.random((50, 21, 21), chunks=(50, 21, 21)) - 1
    x1_phase = solve_cahn_hilliard(x0_phase)
    y0_class = np.zeros(x0_phase.shape[0])
    y1_class = np.ones(x1_phase.shape[0])
    x_combined = np.concatenate((x0_phase, x1_phase))
    y_combined = np.concatenate((y0_class, y1_class))
    homogenization_pipeline.fit(x_combined, y_combined)
    x0_test = 2 * da.random.random((3, 21, 21), chunks=(3, 21, 21)) - 1
    x1_test = solve_cahn_hilliard(x0_test)
    y1_test = homogenization_pipeline.predict(x1_test)
    y0_test = homogenization_pipeline.predict(x0_test)
    assert np.allclose(y0_test, [0, 0, 0])
    assert np.allclose(y1_test, [1, 1, 1])



from .gsh_hex_tri_L0_16 import gsh_basis_info as hex_basis_info
from .gsh_hex_tri_L0_16 import gsh_eval as hex_eval
from .gsh_cub_tri_L0_16 import gsh_basis_info as cub_basis_info
from .gsh_cub_tri_L0_16 import gsh_eval as cub_eval
from .gsh_tri_tri_L0_13 import gsh_basis_info as tri_basis_info
from .gsh_tri_tri_L0_13 import gsh_eval as tri_eval

__all__ = ['hex_basis_info', 'hex_eval', 'cub_basis_info',
           'cub_eval', 'tri_basis_info', 'tri_eval']



import numpy as np
from .fmks.func import deprecate

class Filter(object):
    """
    Wrapper class for convolution with a kernel and resizing of a kernel
    """
    @deprecate
    def __init__(self, Fkernel, basis):
        """
        Instantiate a Filter.

        Args:
            Fkernel: an array representing a convolution kernel
            basis: an instance of a bases class.


        """
        self.basis = basis
        self._Fkernel = Fkernel

    def _frequency_2_real(self, copy=False):
        """
        Converts the kernel from frequency space to real space with
        the origin shifted to the center.

        Args:
            copy (boolean): indicates if _Fkernel should be copied. This is
                used to make sure the influence cofficients remained unchanged.

        Returns:
            an array in real space
        """
        Fkernel = self._Fkernel
        if copy:
            Fkernel = self._Fkernel.copy()
        return np.fft.fftshift(self.basis._ifftn(Fkernel),
                               axes=self.basis._axes)

    def _real_2_frequency(self, kernel):
        """
        Converts a kernel from real space to frequency space.

        Args:
          kernel: an array representing a convolution kernel in real space

        Returns:
          an array in frequency space
        """
        return self.basis._fftn(np.fft.ifftshift(kernel,
                                axes=self.basis._axes))

    def convolve(self, X):
        """
        Convolve X with a kernel in frequency space.

        Args:
          X: array to be convolved

        Returns:
          convolution of X with the kernel
        """
        FX = self.basis._fftn(X)
        if FX.shape[1:] != self._Fkernel.shape[1:]:
            raise RuntimeError("Dimensions of X are incorrect.")
        Fy = self._sum(FX * self._Fkernel)
        return self.basis._ifftn(Fy)

    def _sum(self, Fy):
        return np.sum(Fy, axis=-1)

    def resize(self, size):
        """
        Changes the size of the kernel to size.

        Args:
          size: tuple with the shape of the new kernel
        """
        if len(size) != len(self._Fkernel.shape[1:-1]):
            raise RuntimeError("length of resize shape is incorrect.")
        if not np.all(size >= self._Fkernel.shape[1:-1]):
            raise RuntimeError("resize shape is too small.")
        kernel = self._frequency_2_real()
        kernel_pad = self._zero_pad(kernel, size)
        self._Fkernel = self._real_2_frequency(kernel_pad)
        self.basis._axes_shape = kernel_pad.shape[1:-1]

    def _zero_pad(self, kernel, size):
        """
        Zero pads a real space array with zeros and does a Fourier transform

        Args:
            kernel: real space array
            size: The size of the zero padded array.

        Returns:
            Fourier transform of a zero padded kernel

        """
        if len(size) != kernel.ndim:
            size = kernel.shape[:1] + tuple(size) + kernel.shape[-1:]
        padsize = np.array(size) - np.array(kernel.shape)
        paddown = padsize // 2
        padup = padsize - paddown
        padarray = np.concatenate((padup[..., None],
                                   paddown[..., None]), axis=1)
        pads = tuple([tuple(p) for p in padarray])
        kernel_pad = np.pad(kernel, pads, 'constant', constant_values=0)
        return kernel_pad


class Correlation(Filter):

    """
    Computes the autocorrelation for a microstructure

    >>> n_states = 2
    >>> X = np.array([[[0, 1, 0],
    ...                [0, 1, 0],
    ...                [0, 1, 0]]])
    >>> from pymks.bases import DiscreteIndicatorBasis
    >>> basis = DiscreteIndicatorBasis(n_states=n_states)
    >>> X_ = basis.discretize(X)
    >>> filter_ = Correlation(X_, basis)
    >>> X_auto = filter_.convolve(X_)
    >>> X_test = np.array([[[[3., 0.  ],
    ...                      [6., 3.],
    ...                      [3., 0.  ]],
    ...                     [[3., 0.  ],
    ...                      [6., 3.],
    ...                      [3., 0.  ]],
    ...                     [[3., 0.  ],
    ...                      [6., 3.],
    ...                      [3., 0.  ]]]])
    >>> assert(np.allclose(X_auto, X_test))

    Args:
        X_: The discretized microstructure function, an
            `(n_samples, n_x, ..., n_states)` shaped array
            where `n_samples` is the number of samples, `n_x` is thes
            patial discretization, and n_states is the number of local states.

    Returns:
        Autocorrelations for microstructure X_
    """

    def __init__(self, kernel, basis):
        """
        Instantiate a Correlation.

        Args:
            kernel: an array representing a convolution kernel
            basis: an instance of a bases class.

        """
        self.basis = basis
        if self.basis._axes_shape > kernel.shape[1:-1]:
                kernel = self._zero_pad(kernel, self.basis._axes_shape)
        Fkernel = self.basis._fftn(kernel)
        super(Correlation, self).__init__(np.conjugate(Fkernel), basis)

    def convolve(self, X):
        """
        Convolve X with a kernel in frequency space.

        Args:
            X: array to be convolved

        Returns:
            correlation of X with the kernel
        """
        if X.shape[1:-1] < self.basis._axes_shape:
            X = self._zero_pad(X, self.basis._axes_shape)
        FX = self.basis._fftn(X)
        Fy = self._sum(FX * self._Fkernel)
        correlation = self.basis._ifftn(Fy)
        return np.fft.fftshift(correlation, axes=self.basis._axes)

    def _sum(self, Fy):
        return Fy



"""
The correlation module test cases

"""
import numpy as np
import dask.array as da

from pymks.fmks.correlations import two_point_stats
from pymks.fmks.correlations import correlations_multiple


# pylint: disable=too-many-arguments
def run_one(size, size_predict, chunk, chunks_predict, periodic_boundary, cutoff):
    """Generic function to test two points stats shape and chunks

    """
    x_data = np.random.randint(2, size=size)
    chunks = (chunk,) + x_data.shape[1:]  # pylint: disable=unsubscriptable-object
    x_data = da.from_array(x_data, chunks=chunks)
    stats = two_point_stats(
        x_data, x_data, periodic_boundary=periodic_boundary, cutoff=cutoff
    )
    assert stats.compute().shape == size_predict
    assert stats.chunks == chunks_predict


def test_twodim_odd():
    """This test investigates the stability of the two_point_stats funtion with
    a microstrucure that has odd value height and width
    """
    run_one((2, 15, 15), (2, 9, 9), 1, ((1, 1), (9,), (9,)), False, 4)
    run_one((2, 15, 15), (2, 5, 5), 1, ((1, 1), (5,), (5,)), True, 2)


def test_twodim_even():
    """This test investigates the stability of the two_point_stats funtion with
    a microstrucure that has even value height and width
    """
    run_one((3, 10, 10), (3, 9, 9), 2, ((2, 1), (9,), (9,)), False, 4)
    run_one((3, 10, 10), (3, 5, 5), 2, ((2, 1), (5,), (5,)), True, 2)


def test_twodim_mix1():
    """This test investigates the stability of the two_point_stats funtion with
    a microstrucure that has odd value height and even value width
    """
    run_one((1, 15, 10), (1, 9, 9), 1, ((1,), (9,), (9,)), False, 4)
    run_one((1, 15, 10), (1, 5, 5), 1, ((1,), (5,), (5,)), True, 2)


def test_twodim_mix2():
    """This test investigates the stability of the two_point_stats funtion with
    a microstrucure that has even value height and odd value width
    """
    run_one((1, 10, 15), (1, 9, 9), 1, ((1,), (9,), (9,)), False, 4)
    run_one((1, 10, 15), (1, 5, 5), 1, ((1,), (5,), (5,)), True, 2)


def test_threedim():
    """This test investigates the stability of the two_point_stats funtion with
    3D microstrucure"""
    run_one((4, 10, 10, 15), (4, 9, 9, 9), 2, ((2, 2), (9,), (9,), (9,)), False, 4)
    run_one((4, 10, 10, 15), (4, 5, 5, 5), 2, ((2, 2), (5,), (5,), (5,)), True, 2)


def test_onedim():
    """Test 1D microstructure
    """
    run_one((4, 10), (4, 9), 2, ((2, 2), (9,)), False, 4)
    run_one((4, 10), (4, 5), 2, ((2, 2), (5,)), True, 2)


def test_correlations_multiple():
    """Test that correlations_multiple chunks correctly.

    Test for bug fix. Previously, chunks were ((1, 1), (3,), (3,), (1,
    1)) for this example.

    """
    darr = da.random.randint(2, size=(2, 4, 4, 2), chunks=(1, 4, 4, 2))
    out = correlations_multiple(darr, [[0, 0], [0, 1]])
    assert out.chunks == ((1, 1), (3,), (3,), (2,))



"""Solve the linear elaticity problem to generate data.

Use SfePy to solve a linear strain problem in 2D with a varying
microstructure on a rectangular grid. The rectangle (cube) is held at
the negative edge (plane) and displaced by 1 on the positive x edge
(plane). Periodic boundary conditions are applied to the other
boundaries.

The microstructure is of shape (n_samples, n_x, n_y) or (n_samples,
n_x, n_y, n_z).

>>> X = np.zeros((1, 3, 3), dtype=int)
>>> X[0, :, 1] = 1

>>> strain = _solve_fe(
...     X,
...     elastic_modulus=(1.0, 10.0),
...     poissons_ratio=(0., 0.),
...     macro_strain=1.,
...     delta_x=1.
... )['strain']

y is the strain with components as follows

>>> exx = strain[..., 0]
>>> eyy = strain[..., 1]
>>> exy = strain[..., 2]

In this example, the strain is only in the x-direction and has a
uniform value of 1 since the displacement is always 1 and the size
of the domain is 1.

>>> assert np.allclose(exx, 1)
>>> assert np.allclose(eyy, 0)
>>> assert np.allclose(exy, 0)

The following example is for a system with contrast. It tests the
left/right periodic offset and the top/bottom periodicity.

>>> X = np.array([[[1, 0, 0, 1],
...                [0, 1, 1, 1],
...                [0, 0, 1, 1],
...                [1, 0, 0, 1]]])
>>> n_samples, N, N = X.shape
>>> macro_strain = 0.1
>>> displacement = _solve_fe(
...     X,
...     elastic_modulus=(10.0, 1.0),
...     poissons_ratio=(0.3, 0.3),
...     macro_strain=macro_strain
... )['displacement']
>>> u = displacement[0]

Check that the offset for the left/right planes is `N *
macro_strain`.

>>> assert np.allclose(u[-1,:,0] - u[0,:,0], N * macro_strain)

Check that the left/right side planes are periodic in y.

>>> assert np.allclose(u[0,:,1], u[-1,:,1])

Check that the top/bottom planes are periodic in both x and y.

>>> assert np.allclose(u[:,0], u[:,-1])

The module also works with Dask arrays,

>>> import dask.array as da

>>> X = np.array([[[1, 0, 0, 1],
...                [0, 1, 1, 1],
...                [0, 0, 1, 1],
...                [1, 0, 0, 1]],
...               [[0, 0, 0, 1],
...                [0, 0, 1, 1],
...                [1, 0, 1, 1],
...                [1, 1, 0, 1]]
...              ])
>>> print(X.shape)
(2, 4, 4)
>>> x_data = da.from_array(X, chunks=(1, 4, 4))
>>> out = _solve_fe(x_data,
...             elastic_modulus=(10.0, 1.0),
...             poissons_ratio=(0.3, 0.3),
...             macro_strain=macro_strain)['displacement']
>>> print(out.chunks)
((1, 1), (5,), (5,), (2,))
>>> print(out.shape)
(2, 5, 5, 2)
>>> value = out.compute()
>>> assert np.allclose(value[0, -1, :, 0] - value[0 ,0, :,0], N * macro_strain)

"""

import warnings
import pytest
import numpy as np

from toolz.curried import pipe, do, first, merge
from toolz.curried import map as map_
from toolz.sandbox.parallel import fold

try:
    import sfepy  # pylint: disable=unused-import; # noqa: F401
except ImportError:  # pragma: no cover
    pytest.importorskip("sfepy")
    raise

from sfepy.base.goptions import goptions
from sfepy.discrete.fem import Field

try:
    from sfepy.discrete.fem import FEDomain as Domain
except ImportError:  # pragma: no cover
    from sfepy.discrete.fem import Domain
from sfepy.discrete import (
    FieldVariable,
    Material,
    Integral,
    Function,
    Equation,
    Equations,
    Problem,
)
from sfepy.terms import Term
from sfepy.discrete.conditions import Conditions, EssentialBC, PeriodicBC
from sfepy.solvers.ls import ScipyDirect
from sfepy.solvers.nls import Newton
import sfepy.discrete.fem.periodic as per
from sfepy.discrete import Functions
from sfepy.mesh.mesh_generators import gen_block_mesh
from sfepy.mechanics.matcoefs import ElasticConstants
from sfepy.base.base import output
from sfepy.discrete.conditions import LinearCombinationBC
from sfepy.mechanics.matcoefs import stiffness_from_lame

from ..func import curry, sequence, apply_dict_func

goptions["verbose"] = False
output.set_output(quiet=True)


warnings.simplefilter("ignore", category=FutureWarning)


@curry
def _solve_fe(x_data, elastic_modulus, poissons_ratio, macro_strain=1.0, delta_x=1.0):
    def solve_one_sample(property_array):
        return pipe(
            get_fields(property_array.shape[:-1], delta_x),
            lambda x: get_problem(x, property_array, delta_x, macro_strain),
            lambda x: (x, x.solve()),
            lambda x: get_data(property_array.shape[:-1], *x),
        )

    convert = lambda x: _convert_properties(
        len(x.shape) - 1, elastic_modulus, poissons_ratio
    )[x]

    solve_multiple_samples = sequence(
        do(_check(len(elastic_modulus), len(poissons_ratio))),
        convert,
        map_(solve_one_sample),
        lambda x: zip(*x),
        map_(np.array),
        lambda x: zip(("strain", "displacement", "stress"), tuple(x)),
        dict,
    )

    shape = lambda x: (x.shape[0],) + x.shape[1:] + (3,)
    dis_shape = lambda x: (x.shape[0],) + tuple(y + 1 for y in x.shape[1:]) + (2,)

    if isinstance(x_data, np.ndarray):
        return solve_multiple_samples(x_data)

    return apply_dict_func(
        solve_multiple_samples,
        x_data,
        dict(
            strain=shape(x_data), stress=shape(x_data), displacement=dis_shape(x_data)
        ),
    )


def _convert_properties(dim, elastic_modulus, poissons_ratio):
    """Convert from elastic modulus and Poisson's ratio to the Lame
    parameter and shear modulus

    Args:
      dim: whether 2D or 3D
      elastic_modulus: the elastic modulus in each phase
      poissons_ration: the poissons ratio for each phase

    Returns:
      array of shape (n_phases, 2) where for example [1, 0], gives the
      Lame parameter in the second phase

    >>> assert(np.allclose(
    ...     _convert_properties(
    ...          dim=2, elastic_modulus=(1., 2.), poissons_ratio=(1., 1.)
    ...     ),
    ...     np.array([[-0.5, 1. / 6.], [-1., 1. / 3.]])
    ... ))

    Test case with 3 phases.

    >>> X2D = np.array([[[0, 1, 2, 1],
    ...                  [2, 1, 0, 0],
    ...                  [1, 0, 2, 2]]])
    >>> X2D_property = _convert_properties(
    ...     dim=2, elastic_modulus=(1., 2., 3.), poissons_ratio=(1., 1., 1.)
    ... )[X2D]
    >>> lame = lame0, lame1, lame2 = -0.5, -1., -1.5
    >>> mu = mu0, mu1, mu2 = 1. / 6, 1. / 3, 1. / 2
    >>> lm = list(zip(lame, mu))
    >>> assert(np.allclose(X2D_property,
    ...                    [[lm[0], lm[1], lm[2], lm[1]],
    ...                     [lm[2], lm[1], lm[0], lm[0]],
    ...                     [lm[1], lm[0], lm[2], lm[2]]]))

    Test case with 2 phases.

    >>> X3D = np.array([[[0, 1],
    ...                  [0, 0]],
    ...                 [[1, 1],
    ...                  [0, 1]]])
    >>> X3D_property = _convert_properties(
    ...     dim=2, elastic_modulus=(1., 2.), poissons_ratio=(1., 1.)
    ... )[X3D]

    >>> assert(np.allclose(
    ...     X3D_property,
    ...     [[[lm[0], lm[1]],
    ...       [lm[0], lm[0]]],
    ...      [[lm[1], lm[1]],
    ...       [lm[0], lm[1]]]]
    ... ))

    """
    return pipe(
        zip(elastic_modulus, poissons_ratio),
        map_(
            lambda x: pipe(
                ElasticConstants(young=x[0], poisson=x[1]),
                lambda y: (y.lam, dim / 3.0 * y.mu),
            )
        ),
        list,
        np.array,
    )


@curry
def _check(n_phases, n_phases_other, x_data):
    """Various sanity checks on the data, Elastic modulus and Poissons
    ratio.

    Args:
      n_phases: number of phases in the elastic modulus
      n_phases_other: number of phases in the Poissons ratio
      x_data: the microstructures

    Poissons ratio and elasticy arrays must have equal length

    >>> _check(3, 2, None)
    Traceback (most recent call last):
    ...
    RuntimeError: elastic_modulus and poissons_ratio must be the same length

    The microstructure array must be integer values

    >>> _check(1, 1, np.zeros((2, 2), dtype=float))
    Traceback (most recent call last):
    ...
    TypeError: X must be an integer array

    The microstructure array must be shape (n_sample, n_x, n_y) or
    (n_sample, n_x, n_y, n_z)

    >>> _check(2, 2, np.array([0, 1]).reshape((1, 2)))
    Traceback (most recent call last):
    ...
    RuntimeError: the shape of x_data is incorrect

    The phase index must be bounded by 0 and the n_phases - 1.

    >>> _check(2, 2, np.array([0, 3]))
    Traceback (most recent call last):
    ...
    RuntimeError: X must be between 0 and 1.

    """
    if n_phases != n_phases_other:
        raise RuntimeError("elastic_modulus and poissons_ratio must be the same length")
    if not issubclass(x_data.dtype.type, np.integer):
        raise TypeError("X must be an integer array")
    if np.max(x_data) >= n_phases or np.min(x_data) < 0:
        raise RuntimeError("X must be between 0 and {N}.".format(N=n_phases - 1))
    if not 3 <= len(x_data.shape) <= 4:
        raise RuntimeError("the shape of x_data is incorrect")


def get_fields(shape, delta_x):
    """Get the fields for the displacement and test function

    Args:
      shape: the shape of the domain
      delta_x: the mesh spacing

    Returns:
      tuple of field variables
    """
    return pipe(
        np.array(shape),
        lambda x: gen_block_mesh(
            x * delta_x, x + 1, np.zeros_like(shape), verbose=False
        ),
        lambda x: Domain("domain", x),
        lambda x: x.create_region("region_all", "all"),
        lambda x: Field.from_args("fu", np.float64, "vector", x, approx_order=2),
        lambda x: (
            FieldVariable("u", "unknown", x),
            FieldVariable("v", "test", x, primary_var_name="u"),
        ),
    )


def _get_material(property_array, domain, delta_x):
    """
    Creates an SfePy material from the material property fields for the
    quadrature points.

    Args:
      property_array: array of the properties with shape (n_x, n_y, n_z, 2)
      domain: the Sfepy domain
      delta_x: the grid spacing

    Returns:
      a SfePy material

    """
    reshape = lambda x: np.ascontiguousarray(x.reshape((x.shape[0], 1, 1)))

    def _material_func_(_, coors, mode=None, **__):
        if mode == "qp":
            return pipe(
                np.empty_like(coors, dtype=int),
                lambda x: np.floor(
                    (coors - domain.get_mesh_bounding_box()[0][None]) / delta_x,
                    x,
                    casting="unsafe",
                ),
                lambda x: x.swapaxes(0, 1),
                tuple,
                lambda x: property_array[x],
                lambda x: dict(
                    lam=reshape(x[..., 0]),
                    mu=reshape(x[..., 1]),
                    D=stiffness_from_lame(
                        domain.get_mesh_bounding_box().shape[1],
                        lam=x[..., 0],
                        mu=x[..., 1],
                    ),
                ),
            )
        return None

    return Material("m", function=Function("material_func", _material_func_))


@curry
def get_term(property_array, delta_x, fields):
    """Get the term

    Args:
      property_array: the spatial array of property values
      delta_x: the grid spacing
      fields: the Sfepy u, v fields

    Returns:
      a new term
    """
    return Term.new(
        "dw_lin_elastic_iso(m.lam, m.mu, v, u)",
        Integral("i", order=4),
        fields[0].field.region,
        m=_get_material(property_array, fields[0].field.region.domain, delta_x),
        u=fields[0],
        v=fields[1],
    )


def subdomain_func(x_points=(), y_points=(), z_points=(), max_x=None):
    """
    Creates a function to mask subdomains in Sfepy.

    Args:
      x_points: tuple of lines or points to be masked in the x-plane
      y_points: tuple of lines or points to be masked in the y-plane
      z_points: tuple of lines or points to be masked in the z-plane

    Returns:
      array of masked location indices

    """

    eps = lambda x: 1e-3 * (x[1, -1] - x[0, -1])

    def np_or(seq):
        if seq:
            return fold((lambda x, y: x | y), seq)
        return True

    @curry
    def flag_it(points, coords, index):
        close = lambda x: (coords[:, index] < (x + eps(coords))) & (
            coords[:, index] > (x - eps(coords))
        )
        return pipe(points, map_(close), list, np_or, lambda x: (len(points) == 0) | x)

    def _func(coords, domain=None):  # pylint: disable=unused-argument
        return pipe(
            (x_points, y_points, z_points),
            enumerate,
            map_(lambda x: flag_it(x[1], coords, x[0])),
            list,
            curry(fold)(lambda x, y: x & y),
            lambda x: (x & (coords[:, 0] < (max_x - eps(coords))))
            if max_x is not None
            else x,
            np.where,
            first,
        )

    return _func


@curry
def get_region_func(max_x, dim_string, domain, name_minmax):
    """Generate the Sfepy region

    Args:
      max_x: max value in the x direction
      dim_string: either "x", "y", or "z"
      domain: the Sfepy domain
      name_minmax: tuple of name and min / max

    Returns:
      the Sfepy region
    """
    return pipe(
        subdomain_func(max_x=max_x, **{dim_string + "_points": (name_minmax[1],)}),
        lambda x: Function(dim_string + name_minmax[0], x),
        lambda x: domain.create_region(
            "region_{0}_{1}".format(dim_string, name_minmax[0]),
            "vertices by {0}".format(x.name),
            "facet",
            functions=Functions([x]),
        ),
    )


def get_bc(max_x_func, domain, dim, bc_dict_func):
    """Get the periodic boundary condition

    Args:
      max_x_func: function for finding the maximum value of x
      domain: the Sfepy domain
      dim: the x, y or z direction
      bc_dict_func: function to generate the bc dict

    Returns:
      the boundary condition and the sfepy function
    """
    dim_dict = lambda x: [
        ("x", per.match_x_plane),
        ("y", per.match_y_plane),
        ("z", per.match_z_plane),
    ][dim][x]
    return pipe(
        domain.get_mesh_bounding_box(),
        lambda x: PeriodicBC(
            "periodic_{0}".format(dim_dict(0)),
            list(
                map_(
                    get_region_func(max_x_func(x), dim_dict(0), domain),
                    zip(("plus", "minus"), x[:, dim][::-1]),
                )
            ),
            bc_dict_func(x),
            match="match_{0}_plane".format(dim_dict(0)),
        ),
        lambda x: (x, Function("match_{0}_plane".format(dim_dict(0)), dim_dict(1))),
    )


@curry
def get_periodic_bc_yz(domain, dim):
    """Get the periodic boundary conditoin in the YZ directions

    Args:
      domain: the Sfepy domain
      dim: the x, y or z directions

    Returns:
      the boundary condition and sfepy function
    """

    return get_bc(
        lambda _: None,
        domain,
        dim,
        lambda x: merge({"u.1": "u.1"}, {"u.2": "u.2"} if x.shape[1] == 3 else dict()),
    )


@curry
def get_periodic_bc_x(domain, dim):
    """Get a periodic buondary condition in the X direction

    Args:
      domain: the Sfepy domain
      dim: the x, y or z directions

    Returns:
      the boundary condition and sfepy function
    """
    return get_bc(
        lambda x: domain.get_mesh_bounding_box()[:, 0][1],
        domain,
        dim,
        lambda x: {"u.0": "u.0"},
    )


def get_periodic_bcs(domain):
    """Get the periodic boundary conditions

    Args:
      domain: the Sfepy domain

    Returns:
      the boundary conditions and sfepy functions
    """
    zipped = lambda x, f: pipe(
        range(x, domain.get_mesh_bounding_box().shape[1]),
        map_(f(domain)),
        lambda x: zip(*x),
        list,
    )

    return pipe(
        (zipped(0, get_periodic_bc_yz), zipped(1, get_periodic_bc_x)),
        lambda x: (Conditions(x[0][0] + x[1][0]), Functions(x[0][1] + x[1][1])),
    )


def get_shift_or_fixed_bcs(domain, points_dict_f, name, x_points_f):
    """Generic function for generating fixed or shift boundary conditions

    Args:
      domain: the sfepy domain
      points_dict_f: function to return displacements
      name: the unique of the boundary condition
      x_points_f: function to return x_points for subdomain_func

    Returns:
      a Sfepy EssentialBC
    """

    def func(min_xyz, max_xyz):
        return pipe(
            dict(z_points=(max_xyz[2], min_xyz[2])) if len(min_xyz) == 3 else dict(),
            lambda x: subdomain_func(
                x_points=x_points_f(min_xyz, max_xyz),
                y_points=(max_xyz[1], min_xyz[1]),
                **x,
            ),
            lambda x: Function(f"{name}_x_points", x),
            lambda x: domain.create_region(
                f"region_{name}_points",
                f"vertices by {name}_x_points",
                "vertex",
                functions=Functions([x]),
            ),
            lambda x: EssentialBC(
                f"{name}_points_BC", x, points_dict_f(min_xyz, max_xyz)
            ),
        )

    return func(domain.get_mesh_bounding_box()[0], domain.get_mesh_bounding_box()[1])


def get_displacement_bcs(domain, macro_strain):
    """Get the shift and fixed BCs.

    The shift BC has the the right top and bottom points in x, y and z
    fixed or displaced.

    The fixed BC has the left top and bottom points in x, y and z
    fixed.

    Args:
      domain: an Sfepy domain
      macro_strain: the macro strain

    Returns:
      the Sfepy boundary conditions

    """
    return Conditions(
        [
            get_shift_or_fixed_bcs(
                domain,
                lambda min_, max_: {"u.0": macro_strain * (max_[0] - min_[0])},
                "shift",
                lambda min_, max_: (max_[0],),
            ),
            get_shift_or_fixed_bcs(
                domain,
                lambda min_, max_: merge(
                    {"u.0": 0.0, "u.1": 0.0}, {"u.2": 0.0} if len(min_) == 3 else dict()
                ),
                "fix",
                lambda min_, max_: (min_[0],),
            ),
        ]
    )


def get_linear_combination_bcs(domain, macro_strain):
    """
    The right nodes are periodic with the left nodes but also displaced.

    Args:
      domain: the Sfepy domain
      macro_strain: the macro strain

    Returns:
      linear combination boundary conditions

    """

    def func(min_xyz, max_xyz):
        def shift_(_, coors, __):
            return np.ones_like(coors[:, 0]) * macro_strain * (max_xyz[0] - min_xyz[0])

        return pipe(
            [("plus", max_xyz[0]), ("minus", min_xyz[0])],
            map_(get_region_func(None, "x", domain)),
            list,
            lambda x: LinearCombinationBC(
                "lcbc",
                x,
                {"u.0": "u.0"},
                Function("match_x_plane", per.match_x_plane),
                "shifted_periodic",
                arguments=(Function("shift", shift_),),
            ),
            lambda x: Conditions([x]),
        )

    return func(domain.get_mesh_bounding_box()[0], domain.get_mesh_bounding_box()[1])


def get_nls(evaluator):
    """Get the non-linear solver

    Args:
      evaluator: the problem evaluator

    Returns:
      the non-linear solver
    """
    return Newton(
        {},
        lin_solver=ScipyDirect({}),
        fun=evaluator.eval_residual,
        fun_grad=evaluator.eval_tangent_matrix,
    )


def get_problem(fields, property_array, delta_x, macro_strain):
    """Build the Sfepy problem

    Args:
      fields: the Sfepy fields
      property_array: lame and mu params for every mesh element
      delta_x: the grid spacing
      macro_strain: the macro strain

    Returns:
      the Sfepy problem
    """

    def func(epbcs, functions, ebcs, lcbcs):
        return pipe(
            fields,
            get_term(property_array, delta_x),
            lambda x: [Equation("balance_of_forces", x)],
            Equations,
            lambda x: Problem("elasticity", equations=x, functions=functions),
            do(lambda x: x.time_update(ebcs=ebcs, epbcs=epbcs, lcbcs=lcbcs)),
            do(lambda x: x.set_solver(get_nls(x.get_evaluator()))),
        )

    return pipe(
        fields[0].field.region.domain,
        lambda x: func(
            *get_periodic_bcs(x),
            get_displacement_bcs(x, macro_strain),
            get_linear_combination_bcs(x, macro_strain),
        ),
    )


def get_stress_strain_data(problem, shape, name):
    """Get the stress or strain data

    Args:
      problem: the sovled Sfepy problem
      shape: the shape of the domain
      name: either 'stress' or 'strain' string

    Returns:
      reshaped stress or strain data
    """
    return pipe(
        dict(
            dims=problem.domain.get_mesh_bounding_box().shape[1],
            args=dict(stress="m.D, u", strain="u")[name],
            name=name,
        ),
        lambda x: "ev_cauchy_{name}.{dims}.region_all({args})".format(**x),
        lambda x: problem.evaluate(x, mode="el_avg", copy_materials=False),
        np.squeeze,
        lambda x: np.reshape(x, (shape + x.shape[-1:])),
    )


def get_displacement(vec, shape):
    """Extract the displacement data

    Args:
      vec: the output from problem.solve()
      shape: the grid shape

    Returns:
      displacement field
    """
    return pipe(
        vec.create_output_dict()["u"].data,
        lambda x: np.reshape(x, (tuple(x + 1 for x in shape) + x.shape[-1:])),
    )


@curry
def get_data(shape, problem, vec):
    """Get the data fields given a solved Sfepy problem

    Args:
      shape: the grid shape
      problem: the solved Sfepy problem
      vec: the output from problem.solve()

    Returns
      the reshaped displacement field
    """

    return (
        get_stress_strain_data(problem, shape, "strain"),
        get_displacement(vec, shape),
        get_stress_strain_data(problem, shape, "stress"),
    )



try:
    import matplotlib.pyplot as plt
except ImportError:
    import pytest
    pytest.importorskip('matplotlib')
    raise
import matplotlib.colors as colors
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.axes_grid1 import make_axes_locatable
from sklearn.model_selection import learning_curve
from .stats import _auto_correlations
from .stats import _cross_correlations
import numpy as np
import warnings

from .fmks.func import deprecate
# warnings.filterwarnings("ignore")
plt.style.library['ggplot']['xtick.color'] = '#000000'
plt.style.library['ggplot']['ytick.color'] = '#000000'
plt.style.library['ggplot']['axes.labelcolor'] = '#000000'
plt.style.use('ggplot')



@deprecate
def set_colors():
    """
    Helper function used to set the color map.
    """
    HighRGB = np.array([26, 152, 80]) / 255.
    MediumRGB = np.array([255, 255, 191]) / 255.
    LowRGB = np.array([0, 0, 0]) / 255.
    cdict = _set_cdict(HighRGB, MediumRGB, LowRGB)
    plt.register_cmap(name='PyMKS', data=cdict)
    plt.set_cmap('PyMKS')

@deprecate
def _get_response_cmap():
    """
    Helper function used to set the response color map.

    Returns:
        dictionary with colors and localizations on color bar.
    """
    HighRGB = np.array([179, 255, 204]) / 255.
    MediumRGB = np.array([28, 137, 63]) / 255.
    LowRGB = np.array([11, 53, 24]) / 255.

    cdict = _set_cdict(HighRGB, MediumRGB, LowRGB)
    return colors.LinearSegmentedColormap('coeff_cmap', cdict, 256)

@deprecate
def _get_microstructure_cmap():
    """
    Helper function used to set the microstructure color map.

    Returns:
        dictionary with colors and microstructure on color bar.
    """

    HighRGB = np.array([229, 229, 229]) / 255.
    MediumRGB = np.array([114.5, 114.5, 114.5]) / 255.
    LowRGB = np.array([0, 0, 0]) / 255.
    cdict = _set_cdict(HighRGB, MediumRGB, LowRGB)
    return colors.LinearSegmentedColormap('micro_cmap', cdict, 256)

@deprecate
def _get_diff_cmap():
    """
    Helper function used to set the difference color map.

    Returns:
        dictionary with colors and localizations on color bar.
    """
    HighRGB = np.array([255, 207, 181]) / 255.
    MediumRGB = np.array([238, 86, 52]) / 255.
    LowRGB = np.array([99, 35, 21]) / 255.
    cdict = _set_cdict(HighRGB, MediumRGB, LowRGB)
    return colors.LinearSegmentedColormap('diff_cmap', cdict, 256)

@deprecate
def _grid_matrix_cmap():
    """
    Helper function used to set the grid matrix color map.

    Returns:
        dictionary with colors and localizations on color bar.
    """
    HighRGB = np.array([229, 229, 229]) / 255.
    MediumRGB = np.array([114.5, 114.5, 114.5]) / 255.
    LowRGB = np.array([0, 0, 0]) / 255.
    cdict = _set_cdict(HighRGB, MediumRGB, LowRGB)
    return colors.LinearSegmentedColormap('grid_cmap', cdict, 256)

@deprecate
def _set_cdict(HighRGB, MediumRGB, LowRGB):
    """
    Helper function used to set color map from 3 RGB values.

    Args:
        HighRGB: RGB with highest values
        MediumRGB: RGB with medium values
        LowRGB: RGB with lowest values

    Returns:
        dictionary with colors and localizations on color bar.
    """
    cdict = {'red': ((0.0, LowRGB[0], LowRGB[0]),
                     (0.5, MediumRGB[0], MediumRGB[0]),
                     (1.0, HighRGB[0], HighRGB[0])),

             'green': ((0.0, LowRGB[1], LowRGB[1]),
                       (0.5, MediumRGB[1], MediumRGB[1]),
                       (1.0, HighRGB[1], HighRGB[1])),

             'blue': ((0.0, LowRGB[2], LowRGB[2]),
                      (0.5, MediumRGB[2], MediumRGB[2]),
                      (1.0, HighRGB[2], HighRGB[2]))}

    return cdict

@deprecate
def _get_coeff_cmap():
    """
    Helper function used to set the influence coefficients color map.

    Returns
    """
    HighRGB = np.array([205, 0, 29]) / 255.
    MediumRGB = np.array([240, 240, 240]) / 255.
    LowRGB = np.array([17, 55, 126]) / 255.
    cdict = _set_cdict(HighRGB, MediumRGB, LowRGB)
    return colors.LinearSegmentedColormap('coeff_cmap', cdict, 256)

@deprecate
def _get_color_list(n_sets):
    """
    color list for dimensionality reduction plots

    Args:
        n_sets: number of dataset

    Returns:
        list of colors for n_sets
    """
    color_list = ['#1a9850', '#f46d43', '#1f78b4', '#e31a1c',
                  '#6a3d9a', '#b2df8a', '#fdbf6f', '#a6cee3',
                  '#fb9a99', '#cab2d6', '#ffff99', '#b15928']

    return color_list[:n_sets]

@deprecate
def draw_coeff(coeff, fontsize=15, figsize=None):
    """
    Visualize influence coefficients.

    Args:
        coeff (ND array): influence coefficients with dimensions (x, y,
            n_states)
        fontsize (int, optional): values used for the title font size
    """
    plt.close('all')
    coeff_cmap = _get_coeff_cmap()
    n_coeff = coeff.shape[-1]
    titles = [r'Influence Coefficients $l = %s$' % ii for ii
              in np.arange(n_coeff)]
    _draw_fields(np.rollaxis(coeff, -1, 0), coeff_cmap,
                 fontsize=fontsize, titles=titles, figsize=figsize)

@deprecate
def draw_microstructure_strain(microstructure, strain):
    """
    Draw microstructure and its associated strain

    Args:
        microstructure (2D array): numpy array with dimensions (x, y)
        strain (2D array): numpy array with dimensions (x, y)
    """
    plt.close('all')
    cmap = _get_response_cmap()
    fig = plt.figure(figsize=(8, 4))
    ax0 = plt.subplot(1, 2, 1)
    ax0.imshow(microstructure, cmap=_get_microstructure_cmap(),
               interpolation='none')
    ax0.set_xticks(())
    ax0.set_yticks(())
    ax1 = plt.subplot(1, 2, 2)
    im1 = ax1.imshow(strain, cmap=cmap, interpolation='none')
    ax1.set_xticks(())
    ax1.set_yticks(())
    ax1.set_title(r'$\mathbf{\varepsilon_{xx}}$', fontsize=25)
    ax0.set_title('Microstructure', fontsize=20)
    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([1.0, 0.05, 0.05, 0.9])
    fig.colorbar(im1, cax=cbar_ax)
    plt.tight_layout()
    plt.show()

@deprecate
def draw_microstructures(microstructures, labels=None, figsize=None):
    """
    Draw microstructures

    Args:
        microstructures (3D array): numpy array with dimensions
            (n_samples, x, y)
        labels (list, str, optional): titles for strain fields
        figsize (tuple, optional): specifies the number of images in each
            direction.
    """
    cmap = _get_microstructure_cmap()
    if labels is None:
        labels = [' ' for s in np.arange(microstructures.shape[0])]
    _draw_fields(microstructures, cmap, 15, labels, figsize=figsize)

@deprecate
def draw_strains(strains, labels=None, fontsize=15):
    """
    Draw strain fields

    Args:
        strains (3D array): numpy arrays with dimensions (n_samples, x, y)
        labels (list, str, optional): titles for strain fields
        fontsize (int, optional): title font size
    """
    cmap = _get_response_cmap()
    if labels is None:
        labels = [' ' for s in strains]
    _draw_fields(strains, cmap, fontsize, labels)

@deprecate
def draw_concentrations(concentrations, labels=None, fontsize=15):
    """Draw comparison fields

    Args:
        concentrations (list): numpy arrays with dimensions (x, y)
        labels (list): titles for concentrations
        fontsize (int): used for the title font size
    """
    if labels is None:
        labels = [" " for s in concentrations]
    cmap = _get_response_cmap()
    _draw_fields(concentrations, cmap, fontsize, labels)

@deprecate
def draw_strains_compare(strain_FEM, strain_MKS, fontsize=20):
    """Draw comparison of strain fields.

    Args:
        strain_FEM (2D array): strain field with dimensions (x, y) from finite
            element
        strain_MKS (2D array): strain fieldwith dimensions (x, y) from MKS
        fontsize (int, optional): scalar values used for the title font size
    """
    cmap = _get_response_cmap()
    titles = ['Finite Element', 'MKS']
    titles_ = [r'$\mathbf{\varepsilon_{xx}}$ - %s' % title for title in titles]
    _draw_fields((strain_FEM, strain_MKS), cmap, fontsize, titles_)

@deprecate
def draw_concentrations_compare(concentrations, labels, fontsize=15):
    """Draw comparesion of concentrations.

    Args:
        concentrations (3D array): list of difference arrays with dimensions
            (x, y)
        labels (list, str): list of titles for difference arrays
        fontsize (int, optional): scalar values used for the title font size
    """
    cmap = _get_response_cmap()
    _draw_fields(concentrations, cmap, fontsize, labels)

@deprecate
def draw_differences(differences, labels=None, fontsize=15):
    """Draw differences in predicted response fields.

    Args:
        differences (list, 2D arrays): list of difference arrays with
            dimesions (x, y).
        labels (list, str, optional): titles for difference arrays
        fontsize (int, optional): scalar values used for the title font size
    """
    cmap = _get_diff_cmap()
    if labels is None:
        labels = [' ' for s in differences]
    _draw_fields(differences, cmap, fontsize, labels)

@deprecate
def _draw_fields(fields, field_cmap, fontsize, titles, figsize=None):
    """
    Helper function used to draw fields.

    Args:
        fields - iterable object with 2D numpy arrays
        field_cmap - color map for plot
        fontsize - font size for titles and color bar text
        titles - titles for plot
        figsize - controls the number of images in each direction
    """
    plt.close('all')
    vmin = np.min(fields)
    vmax = np.max(fields)
    n_fields = len(fields)
    if titles is not None:
        n_titles = len(titles)
        if n_fields != n_titles:
            raise RuntimeError(
                "number of plots does not match number of labels.")
    plt.close('all')
    if figsize is None:
        figsize = (1, n_fields)
    fig, axs = plt.subplots(figsize[0], figsize[1],
                            figsize=(figsize[1] * 4, figsize[0] * 4))

    if n_fields > 1:
        for field, ax, title in zip(fields, axs.flat, titles):
            im = ax.imshow(field,
                           cmap=field_cmap, interpolation='none',
                           vmin=vmin, vmax=vmax)
            ax.set_xticks(())
            ax.set_yticks(())
            ax.set_title(title, fontsize=fontsize)
    else:
        im = axs.imshow(fields[0], cmap=field_cmap,
                        interpolation='none', vmin=vmin, vmax=vmax)
        axs.set_xticks(())
        axs.set_yticks(())
        axs.set_title(titles[0], fontsize=fontsize)
    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([1.0, 0.05, 0.05, 0.9])
    cbar_font = np.floor(0.8 * fontsize)
    cbar_ax.tick_params(labelsize=cbar_font)
    cbar_ax.yaxis.set_offset_position('right')
    fig.colorbar(im, cax=cbar_ax)
    plt.rc('font', **{'size': str(cbar_font)})
    plt.tight_layout()
    plt.show()

@deprecate
def draw_gridscores(grid_scores, param, score_label=None, colors=None,
                    data_labels=None, param_label=None, fontsize=20):
    """
    Visualize the score values and standard deviations from grids
    scores result from GridSearchCV while varying 1 parameters.

    Args:
        grid_scores (list, grid_scores): `grid_scores_` attribute from
            GridSearchCV
        param (list, str): parameters used in grid_scores
        score_label (str): label for score value axis
        colors (list): colors used for this specified parameter
        param_label (list): parameter titles to appear on plot
    """
    plt.close('all')
    if type(grid_scores[0]) is not list:
        grid_scores = [grid_scores]
    if data_labels is None:
        data_labels = [None for l in range(len(grid_scores))]
    if score_label is None:
        score_label = ''
    if param_label is None:
        param_label is ''
    if colors is None:
        colors = _get_color_list(len(grid_scores))
    if len(grid_scores) != len(data_labels) or len(data_labels) != len(colors):
        raise RuntimeError(
            "grid_scores, colors, and param_lables must have the same length.")
    mins, maxes = [], []
    for grid_score, data_label, color in zip(grid_scores, data_labels, colors):
        tmp = [[item['params'][param], item['mean_test_score'], item['std_test_score']] for item in grid_score]
        _param, errors, stddev = list(zip(*tmp))
        _mins = np.array(errors) - np.array(stddev)
        _maxes = np.array(errors) + np.array(stddev)
        plt.fill_between(_param, _mins, _maxes, alpha=0.1,
                         color=color)
        plt.plot(_param, errors, 'o-', color=color, label=data_label,
                 linewidth=2)
        mins.append(min(_mins))
        maxes.append(max(_maxes))
    if data_labels[0] is not None:
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.,
                   fontsize=15)
    _min, _max = min(mins), max(maxes)
    y_epsilon = (_max - _min) * 0.05
    plt.ylim((_min - y_epsilon, _max + y_epsilon))
    plt.ticklabel_format(style='sci', axis='y')
    plt.ylabel(score_label, fontsize=fontsize)
    plt.xlabel(param_label, fontsize=fontsize)
    plt.show()

@deprecate
def draw_gridscores_matrix(grid_scores, params, score_label=None,
                           param_labels=None):
    """
    Visualize the score value matrix and standard deviation matrix from grids
    scores result from GridSearchCV while varying two parameters.

    Args:
        grid_scores (list): `grid_scores_` attribute from GridSearchCV
        params (list): two parameters used in grid_scores
        score_label (str): label for score value axis
        param_labels (list): parameter titles to appear on plot
    """
    plt.close('all')
    if score_label is None:
        score_label = 'R-Squared'
    if param_labels is None:
        param_labels = ['', '']

    param = grid_scores.cv_results_['params']
    means = grid_scores.cv_results_['mean_test_score']
    stddev = grid_scores.cv_results_['std_test_score']

#    tmp = [[params, mean_score, scores.std()]
#           for parameters, mean_score, scores in grid_scores.grid_scores_]
#    param, means, stddev = list(zip(*tmp))
    param_range_0 = grid_scores.param_grid[params[0]]
    param_range_1 = grid_scores.param_grid[params[1]]
    mat_size = (len(param_range_1), len(param_range_0))
    fig, axs = plt.subplots(2, 1, figsize=(10, 5))
    matrices = np.concatenate((np.array(means).reshape(mat_size)[None],
                               np.array(stddev).reshape(mat_size)[None]))
    X_cmap = _grid_matrix_cmap()
    x_label = param_labels[0]
    y_label = param_labels[1]
    plot_title = [score_label, 'Standard Deviation']
    for ax, label, matrix, title in zip(axs, param_labels,
                                        matrices,
                                        plot_title):
        ax.set_xticklabels(param_range_0, fontsize=12)
        ax.set_yticklabels(param_range_1, fontsize=12)
        ax.set_xticks(np.arange(len(param_range_0)))
        ax.set_yticks(np.arange(len(param_range_0)))
        ax.set_xlabel(x_label, fontsize=14)
        ax.set_ylabel(y_label, fontsize=14)
        ax.grid(False)
        im = ax.imshow(matrix,
                       cmap=X_cmap, interpolation='none')
        ax.set_title(title, fontsize=22)
        divider = make_axes_locatable(ax)
        cbar_ax = divider.append_axes("right", size="10%", pad=0.05)
        cbar = plt.colorbar(im, cax=cbar_ax)
        cbar.ax.tick_params(labelsize=12)
        fig.subplots_adjust(right=1.2)
    plt.tight_layout()
    plt.show()

@deprecate
def draw_component_variance(variance):
    """
    Visualize the percent variance as a function of components.

    Args:
        variance (list): variance ratio explanation from dimensional
            reduction technique.
    """
    plt.close('all')
    n_components = len(variance)
    x = np.arange(1, n_components + 1)
    plt.plot(x, np.cumsum(variance * 100), 'o-', color='#1a9641', linewidth=2)
    plt.xlabel('Number of Components', fontsize=15)
    plt.xlim(0, n_components + 1)
    plt.ylabel('Percent Variance', fontsize=15)
    plt.show()

@deprecate
def draw_components_scatter(datasets, labels, title=None,
                            component_labels=None, view_angles=None,
                            legend_outside=False, fig_size=None):
    """
    Visualize low dimensional representations of microstructures.

    Args:
        datasets (list, 2D arrays): low dimensional data with dimensions
            [n_samples, n_components]. The length of n_components must be 2 or
            3.
        labels (list, str): list of lables for each of each array datasets
        title: main title for plot
        component_labels: labels for components
        view_angles (int,int): the elevation and azimuth angles of the axes
            to rotate the axes.
        legend_outside: specify to move legend box outside the main plot
            domain
        figsize: (width, height) figure size in inches
    """
    plt.close('all')
    if title is None:
        title = 'Low Dimensional Representation'
    n_components = np.array(datasets[0][-1].shape, dtype=int)
    if component_labels is None:
        component_labels = range(1, n_components[0] + 1)
    if len(datasets) != len(labels):
        raise RuntimeError('datasets and labels must have the same length')
    if n_components != len(component_labels):
        raise RuntimeError('number of components and component_labels must'
                           ' have the same length')
    if n_components[-1] == 2:
        _draw_components_2D(datasets, labels, title, component_labels[:2],
                            legend_outside, fig_size)
    elif n_components[-1] == 3:
        _draw_components_3D(datasets, labels, title, component_labels,
                            view_angles, legend_outside, fig_size)
    else:
        raise RuntimeError("n_components must be 2 or 3.")

@deprecate
def draw_evolution(datasets, labels, title=None, component_labels=None,
                   view_angles=None, legend_outside=False, fig_size=None):
    """
    Visualize low dimensional representations of microstructures.

    Args:
        datasets (list, 2D arrays): low dimensional data with dimensions
            [n_samples, n_components]. The length of n_components must be 2 or
            3.
        labels (list, str): list of lables for each of each array datasets
        title: main title for plot
        component_labels: labels for components
        view_angles (int,int): the elevation and azimuth angles of the axes
            to rotate the axes.
        legend_outside: specify to move legend box outside the main plot
            domain
        figsize: (width, height) figure size in inches
    """
    plt.close('all')
    if title is None:
        title = 'Low Dimensional Representation'
    n_components = np.array(datasets[0][-1].shape)
    if component_labels is None:
        component_labels = range(1, n_components + 1)
    if len(datasets) != len(labels):
        raise RuntimeError('datasets and labels must have the same length')
    if n_components != len(component_labels):
        raise RuntimeError('number of components and component_labels must'
                           ' have the same length')
    if n_components[-1] == 2:
        _draw_components_evolution(datasets, labels,
                                   title, component_labels[:2],
                                   legend_outside, fig_size)
    else:
        raise RuntimeError("time and one component must be paired")

@deprecate
def _draw_components_2D(X, labels, title, component_labels,
                        legend_outside, fig_size):
    """
    Helper function to plot 2 components.

    Args:
        X: Arrays with low dimensional data
        labels: labels for each of the low dimensional arrays
    """
    n_sets = len(X)
    color_list = _get_color_list(n_sets)
    if fig_size is not None:
        fig = plt.figure(figsize=(fig_size[0], fig_size[1]))
    else:
        fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_xlabel('Component ' + str(component_labels[0]), fontsize=15)
    ax.set_ylabel('Component ' + str(component_labels[1]), fontsize=15)
    X_array = np.concatenate(X)
    x_min, x_max = [np.min(X_array[:, 0]), np.max(X_array[:, 0])]
    y_min, y_max = [np.min(X_array[:, 1]), np.max(X_array[:, 1])]
    x_epsilon = (x_max - x_min) * 0.05
    y_epsilon = (y_max - y_min) * 0.05
    ax.set_xlim([x_min - x_epsilon, x_max + x_epsilon])
    ax.set_ylim([y_min - y_epsilon, y_max + y_epsilon])
    for label, pts, color in zip(labels, X, color_list):
        ax.plot(pts[:, 0], pts[:, 1], 'o', color=color, label=label)
        lg = plt.legend(loc=1, borderaxespad=0., fontsize=15)
    if legend_outside is not False:
        lg = plt.legend(bbox_to_anchor=(1.05, 1.0), loc=2,
                        borderaxespad=0., fontsize=15)
    plt.title(title, fontsize=20)
    plt.show()

@deprecate
def _draw_components_evolution(X, labels, title, component_labels,
                               legend_outside, fig_size):
    """
    Helper function to plot 2 components.

    Args:
        X: Arrays with low dimensional data
        labels: labels for each of the low dimensional arrays
    """
    n_sets = len(X)
    color_list = _get_color_list(n_sets)
    if fig_size is not None:
        fig = plt.figure(figsize=(fig_size[0], fig_size[1]))
    else:
        fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_xlabel('Time', fontsize=15)
    ax.set_ylabel('Components ', fontsize=15)
    X_array = np.concatenate(X)
    x_min, x_max = [np.min(X_array[:, 0]), np.max(X_array[:, 0])]
    y_min, y_max = [np.min(X_array[:, 1]), np.max(X_array[:, 1])]
    x_epsilon = (x_max - x_min) * 0.05
    y_epsilon = (y_max - y_min) * 0.05
    ax.set_xlim([x_min - x_epsilon, x_max + x_epsilon])
    ax.set_ylim([y_min - y_epsilon, y_max + y_epsilon])
    for label, pts, color in zip(labels, X, color_list):
        ax.plot(pts[:, 0], pts[:, 1], 'o', color=color, label=label)
        lg = plt.legend(loc=1, borderaxespad=0., fontsize=15)
    if legend_outside:
        lg = plt.legend(bbox_to_anchor=(1.05, 1.0), loc=2,
                        borderaxespad=0., fontsize=15)
    plt.title(title, fontsize=20)
    plt.show()

@deprecate
def _draw_components_3D(X, labels, title, component_labels, view_angles,
                        legend_outside, fig_size):
    """
    Helper function to plot 2 components.

    Args:
        X: Arrays with low dimensional data
        labels: labels for each of the low dimensional arrays
    """
    n_sets = len(X)
    color_list = _get_color_list(n_sets)
    if fig_size is not None:
        fig = plt.figure(figsize=(fig_size[0], fig_size[1]))
    else:
        fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.set_xlabel('Component ' + str(component_labels[0]), fontsize=12)
    ax.set_ylabel('Component ' + str(component_labels[1]), fontsize=12)
    ax.set_zlabel('Component ' + str(component_labels[2]), fontsize=12)
    X_array = np.concatenate(X)
    x_min, x_max = [np.min(X_array[:, 0]), np.max(X_array[:, 0])]
    y_min, y_max = [np.min(X_array[:, 1]), np.max(X_array[:, 1])]
    z_min, z_max = [np.min(X_array[:, 2]), np.max(X_array[:, 2])]
    x_epsilon = (x_max - x_min) * 0.05
    y_epsilon = (y_max - y_min) * 0.05
    z_epsilon = (z_max - z_min) * 0.05
    ax.set_xlim([x_min - x_epsilon, x_max + x_epsilon])
    ax.set_ylim([y_min - y_epsilon, y_max + y_epsilon])
    ax.set_zlim([z_min - z_epsilon, z_max + z_epsilon])
    for label, pts, color in zip(labels, X, color_list):
        ax.plot(pts[:, 0], pts[:, 1], pts[:, 2], 'o', color=color, label=label)
    plt.title(title, fontsize=15)
    if view_angles is not None:
        ax.view_init(view_angles[0], view_angles[1])
    lg = plt.legend(loc=1, borderaxespad=0., fontsize=15)
    if legend_outside:
        lg = plt.legend(bbox_to_anchor=(1.05, 1.0), loc=2,
                        borderaxespad=0., fontsize=15)
    plt.show()

@deprecate
def draw_goodness_of_fit(fit_data, pred_data, labels):
    """Goodness of fit plot for MKSHomogenizationModel.

    Args:
        fit_data (2D array): Low dimensional representation of the prediction
            values of the data used to fit the model and the actual values.
        pred_data (2D array): Low dimensional representation of the prediction
            values of the data used for prediction with the model and the
            actual values.
    """
    plt.close('all')
    y_total = np.concatenate((fit_data, pred_data), axis=-1)
    y_min, y_max = np.min(y_total), np.max(y_total)
    middle = (y_max + y_min) / 2.
    data_range = y_max - y_min
    line = np.linspace(middle - data_range * 1.03 / 2,
                       middle + data_range * 1.03 / 2, endpoint=False)
    plt.plot(line, line, '-', linewidth=3, color='#000000')
    plt.plot(fit_data[0], fit_data[1], 'o', color='#1a9850', label=labels[0])
    plt.plot(pred_data[0], pred_data[1], 'o',
             color='#f46d43', label=labels[1])
    plt.title('Goodness of Fit', fontsize=20)
    plt.xlabel('Actual', fontsize=18)
    plt.ylabel('Predicted', fontsize=18)
    plt.legend(loc=2, fontsize=15)
    plt.show()

@deprecate
def draw_components(X_comp, fontsize=15, figsize=None):
    """
    Visualize spatial correlations.

    Args:
        X_corr (ND array): correlations
        correlations (list, optional): correlation labels
    """
    cmap = _get_coeff_cmap()
    titles = [r'Component $%s$' % (ii + 1) for ii
              in np.arange(X_comp.shape[0])]
    _draw_fields(X_comp, cmap, fontsize, titles, figsize=figsize)

@deprecate
def draw_correlations(X_corr, correlations=None):
    """
    Visualize spatial correlations.

    Args:
        X_corr (ND array): correlations
        correlations (list, optional): correlation labels
    """
    if correlations is None:
        n_cross = X_corr.shape[-1]
        L = range(int((np.sqrt(1 + 8 * n_cross) - 1) / 2))
        correlations = list(zip(*list(_auto_correlations(L))))
        correlations += list(zip(*list(_cross_correlations(L))))
    _draw_stats(X_corr, correlations=correlations)

@deprecate
def draw_autocorrelations(X_auto, autocorrelations=None):
    """
    Visualize spatial autocorrelations.

    Args:
        X_auto (ND array): autocorrelations
        autocorrelations (list, optional): autocorrelation labels.
    """
    if autocorrelations is None:
        n_states = X_auto.shape[-1]
        autocorrelations = zip(*list(_auto_correlations(n_states)))
    _draw_stats(X_auto, correlations=autocorrelations)

@deprecate
def draw_crosscorrelations(X_cross, crosscorrelations=None):
    """
    Visualize spatial crosscorrelations.

    Args:
        X_cross (ND array): cross-correlations
        correlations (list, optional): cross-correlation labels.
    """
    if crosscorrelations is None:
        n_cross = X_cross.shape[-1]
        n_states = (np.sqrt(1 + 8 * n_cross) + 1).astype(int) // 2
        crosscorrelations = zip(*list(_cross_correlations(n_states)))
    _draw_stats(X_cross, correlations=crosscorrelations)

@deprecate
def _draw_stats(X_, correlations=None):
    """Visualize the spatial correlations.

    Args:
        X_: correlations
        correlations: list of tuples to select the spatial correlations
            that will be displayed.
    """
    plt.close('all')
    X_cmap = _get_coeff_cmap()
    n_plots = len(correlations)
    x_loc, x_labels = _get_ticks_params(X_.shape[0])
    y_loc, y_labels = _get_ticks_params(X_.shape[1])
    fig, axs = plt.subplots(1, n_plots, figsize=(n_plots * 5, 5))
    if n_plots == 1:
        axs = list([axs])
    for ax, label, img in zip(axs, correlations, np.rollaxis(X_, -1)):
        ax.grid(False)
        ax.set_xticks(x_loc)
        ax.set_xticklabels(x_labels, fontsize=12)
        ax.set_yticks(y_loc)
        ax.set_yticklabels(y_labels, fontsize=12)
        im = ax.imshow(img, cmap=X_cmap, interpolation='none')
        ax.set_title(r"Correlation $l = {0}$, $l' = {1}$".format(label[0],
                                                                 label[1]),
                     fontsize=15)
        fig.subplots_adjust(right=0.8)
        divider = make_axes_locatable(ax)

        cbar_ax = divider.append_axes("right", size="10%", pad=0.05)
        cbar_ticks = _get_colorbar_ticks(img, 5)
        cbar_ticks_diff = cbar_ticks[-1] - cbar_ticks[0]
        cbar_top, cbar_grids = np.max(X_) * 0.005, 0.005
        if cbar_ticks_diff <= 1e-15:
            cbar_top = 0.
            cbar_grids = 0.5
        try:
            cbar = plt.colorbar(im, cax=cbar_ax, ticks=cbar_ticks,
                                boundaries=np.arange(cbar_ticks[0],
                                                     cbar_ticks[-1] + cbar_top,
                                                     cbar_ticks_diff *
                                                     cbar_grids))
            cbar.ax.tick_params(labelsize=12)
        except:
            cbar = plt.colorbar(im, cax=cbar_ax, boundaries=np.unique(X_))
        cbar.ax.tick_params(labelsize=12)
        fig.subplots_adjust(right=0.8)
        plt.tight_layout()
    plt.show()

@deprecate
def _get_ticks_params(l):
    """Get tick locations and labels for spatial correlation plots.

    Args:
        l: shape of array along the axis
    """
    segments = np.roll(np.arange(4, 7, dtype=int), 1, 0)
    m = segments[np.argmin(l % segments)]
    n = int(max((l + 1) / m, 1))
    tick_loc = list(range(0, l + n, n))
    tick_labels = list(range(int(round(- (l - 1) / 2)),
                       int(round(int((l + 1) / 2 + n))), n))
    return tick_loc, tick_labels

@deprecate
def _get_colorbar_ticks(X_, n_ticks):
    """
    Helper function to get colorbar color tick locations.

    Args:
        X: sspatial correlations array
           (n_samples, x,  y, local_state_correlation)
    """
    tick_range = np.linspace(np.min(X_), np.max(X_), n_ticks)
    return tick_range.astype(float) # pylint: disable=no-member

@deprecate
def draw_learning_curves(estimator, X, y, ylim=None, cv=None, n_jobs=1,
                         scoring=None, train_sizes=np.linspace(.1, 1.0, 5)):
    """Code taken from scikit-learn examples for version 0.15.

    Generate a simple plot of the test and traning learning curve.

    Args:
        estimator (class): object type that implements the "fit" and "predict"
            methods
            An object of that type which is cloned for each validation.
        title (str): Used for the title for the chart.
        X (2D array): array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.
        y (1D array): array-like, shape (n_samples) or (n_samples,
            n_features), optional Target relative to X for classification or
            regression; None for unsupervised learning.
        ylim (tuple, optional): Defines minimum and maximum yvalues plotted.
        cv (int, optional): If an integer is passed, it is the number of folds
            (defaults to 3). Specific cross-validation objects can be passed,
            see sklearn.cross_validation module for the list of possible
            objects
        n_jobs(int, optional) : Number of jobs to run in parallel (default 1).
        train_sizes (float): Relative or absolute numbers of training examples
            that will be used to generate the learning curve. If the dtype is
            float, it is regarded as a fraction of the maximum size of the
            training set (that is determined by the selected validation
            method), i.e. it has to be within (0, 1]. Otherwise it is
            interpreted as absolute sizes of the training sets. Note that for
            classification the number of samples usually have to be big enough
            to contain at least one sample from each class. (default:
            np.linspace(0.1, 1.0, 5))

        Returns:
            A plot of the learning curves for both the training curve and the
            cross-validation curve.
    """
    plt.close('all')
    flat_shape = (X.shape[0],) + (np.prod(X.shape[1:]),)
    X_flat = X.reshape(flat_shape)
    plt.figure()
    plt.title('Learning Curves', fontsize=20)
    if ylim is not None:
        plt.ylim(*ylim)
    plt.xlabel("Training examples", fontsize=15)
    plt.ylabel("Score", fontsize=15)
    train_sizes, train_scores, test_scores = learning_curve(
        estimator, X_flat, y, cv=cv, n_jobs=n_jobs,
        train_sizes=train_sizes, scoring=scoring)
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    plt.grid()

    plt.fill_between(train_sizes, train_scores_mean - train_scores_std,
                     train_scores_mean + train_scores_std, alpha=0.1,
                     color="#f46d43")
    plt.fill_between(train_sizes, test_scores_mean - test_scores_std,
                     test_scores_mean + test_scores_std, alpha=0.1,
                     color="#1a9641")
    plt.plot(train_sizes, train_scores_mean, 'o-', color="#f46d43",
             linewidth=2, label="Training score")
    plt.plot(train_sizes, test_scores_mean, 'o-', color="#1a9641",
             linewidth=2, label="Cross-validation score")

    plt.legend(loc="best")
    plt.show()



from .abstract import _AbstractMicrostructureBasis
from .fftmodule import rfftn, irfftn


class _RealFFTBasis(_AbstractMicrostructureBasis):
    """This class is used to make the bases that create real valued
    microstructure functions use the real rFFT/irFFT algorithms and selects
    the appropriate fft module depending on whether or not pyfftw is installed.
    """

    def _fftn(self, X):
        """Real rFFT algorithm

        Args:
            X: NDarray (n_samples, N_x, ...)

        Returns:
            Fourier transform of X
        """
        return rfftn(X, axes=self._axes, threads=self._n_jobs)

    def _ifftn(self, X):
        """Real irFFT algorithm

        Args:
            X: NDarray (n_samples, N_x, ...)

        Returns:
            Inverse Fourier transform of X
        """
        return irfftn(X, axes_shape=self._axes_shape, axes=self._axes, threads=self._n_jobs).real



"""Simple filter example
"""

import numpy as np

from toolz import pipe
import dask.array as da

from pymks.fmks.func import dafft, daifft
from pymks.fmks.localization import fit, coeff_to_real
from pymks.fmks.bases.primitive import discretize, redundancy


def _filter(data):
    return np.where(
        data < 10,
        np.exp(-abs(data)) * np.cos(data * np.pi),
        np.exp(-abs(data - 20)) * np.cos((data - 20) * np.pi),
    )


def _coeff(n_space, n_state):
    return (
        np.linspace(1, 0, n_state)[None, :]
        * _filter(np.linspace(0, 20, n_space))[:, None]
    )


def _fcoeff(n_space, n_state):
    return np.fft.fft(_coeff(n_space, n_state), axis=0)


def _response(x_data, n_space, n_state):
    return pipe(
        np.linspace(0, 1, n_state),
        lambda h: da.maximum(1 - abs(x_data[:, :, None] - h) / (h[1] - h[0]), 0),
        dafft(axis=1),
        lambda fx: da.sum(_fcoeff(n_space, n_state)[None] * fx, axis=-1),
        daifft(axis=1),
    ).real


def _mks_fcoeff(x_data, n_space, n_state):
    return fit(
        x_data,
        _response(x_data, n_space, n_state),
        discretize(n_state=n_state),
        redundancy_func=redundancy,
    )


def _mks_coeff(x_data, n_space, n_state):
    return pipe(
        _mks_fcoeff(x_data, n_space, n_state), coeff_to_real(new_shape=(n_space,))
    )


def _calc_coeff(n_space, n_state):
    return np.fft.fftshift(_coeff(n_space, n_state), axes=(0,))


def _compare(n_sample, n_space, n_state, chunk):
    np.random.seed(2)
    x_data = da.from_array(
        np.random.random((n_sample, n_space)), chunks=(chunk, n_space)
    )
    return np.allclose(
        _calc_coeff(n_space, n_state), _mks_coeff(x_data, n_space, n_state)
    )


def test():
    """Test a simple filter example.
    """
    assert _compare(400, 81, 2, 400)



import numpy as np
from test import get_delta_data, get_random_data


def test_MKS_elastic_delta():
    from pymks import MKSRegressionModel
    from pymks.bases import DiscreteIndicatorBasis
    nx, ny = 21, 21
    X, y_test = get_delta_data(nx, ny)
    basis = DiscreteIndicatorBasis(n_states=2)
    model = MKSRegressionModel(basis)
    model.fit(X, y_test)
    y_pred = model.predict(X)
    assert np.allclose(y_pred, y_test, rtol=1e-3, atol=1e-3)


def test_MKS_elastic_random():
    from pymks import MKSRegressionModel
    from pymks.bases import DiscreteIndicatorBasis
    nx, ny = 21, 21
    X_delta, y_delta = get_delta_data(nx, ny)
    X_test, y_test = get_random_data(nx, ny)
    basis = DiscreteIndicatorBasis(n_states=2)
    model = MKSRegressionModel(basis)
    model.fit(X_delta, y_delta)
    y_pred = model.predict(X_test)
    assert np.allclose(y_pred, y_test, rtol=1e-2, atol=6.1e-3)


def test_resize_pred():
    from pymks import MKSRegressionModel
    from pymks.bases import DiscreteIndicatorBasis

    nx, ny = 21, 21
    resize = 3
    X_delta, y_delta = get_delta_data(nx, ny)
    X_test, y_test = get_random_data(nx, ny)
    X_big_test, y_big_test = get_random_data(resize * nx, resize * ny)
    basis = DiscreteIndicatorBasis(n_states=2)

    model = MKSRegressionModel(basis)
    model.fit(X_delta, y_delta)
    y_pred = model.predict(X_test)
    assert np.allclose(y_pred, y_test, rtol=1e-2, atol=6.1e-3)
    model.resize_coeff((resize * nx, resize * ny))
    y_big_pred = model.predict(X_big_test)
    assert np.allclose(y_big_pred, y_big_test, rtol=1e-2, atol=6.1e-2)


def test_resize_coeff():
    from pymks import MKSRegressionModel
    from pymks.bases import DiscreteIndicatorBasis

    nx, ny = 21, 21
    resize = 3
    X_delta, y_delta = get_delta_data(nx, ny)
    X_big_delta, y_big_delta = get_delta_data(resize * nx, resize * ny)
    basis = DiscreteIndicatorBasis(n_states=2)
    model = MKSRegressionModel(basis)
    big_model = MKSRegressionModel(basis)
    model.fit(X_delta, y_delta)
    big_model.fit(X_big_delta, y_big_delta)
    model.resize_coeff((resize * nx, resize * ny))
    assert np.allclose(model.coef_, big_model.coef_,
                       rtol=1e-2, atol=2.1e-3)


def test_multiphase_FE_strain():
    from pymks import MKSRegressionModel
    from pymks.datasets import make_elastic_FE_strain_delta
    from pymks.datasets import make_elastic_FE_strain_random
    from pymks.bases import DiscreteIndicatorBasis

    L = 21
    i = 3
    elastic_modulus = (80, 100, 120)
    poissons_ratio = (0.3, 0.3, 0.3)
    macro_strain = 0.02
    size = (L, L)

    X_delta, strains_delta = \
        make_elastic_FE_strain_delta(elastic_modulus=elastic_modulus,
                                     poissons_ratio=poissons_ratio,
                                     size=size, macro_strain=macro_strain)
    basis = DiscreteIndicatorBasis(len(elastic_modulus))
    MKSmodel = MKSRegressionModel(basis)
    MKSmodel.fit(X_delta, strains_delta)
    np.random.seed(99)
    X, strain = make_elastic_FE_strain_random(n_samples=1,
                                              elastic_modulus=elastic_modulus,
                                              poissons_ratio=poissons_ratio,
                                              size=size,
                                              macro_strain=macro_strain)
    strain_pred = MKSmodel.predict(X)

    assert np.allclose(strain_pred[0, i:-i], strain[0, i:-i],
                       rtol=1e-2, atol=6.1e-3)


def test_coeff_stablity_with_irfftn():
    from pymks import MKSRegressionModel
    from pymks.bases import DiscreteIndicatorBasis

    nx, ny = 21, 21
    resize = 3
    X_delta, y_delta = get_delta_data(nx, ny)
    X_test, y_test = get_random_data(nx, ny)
    X_big_test, y_big_test = get_random_data(resize * nx, resize * ny)
    basis = DiscreteIndicatorBasis(n_states=2)

    model = MKSRegressionModel(basis)
    model.fit(X_delta, y_delta)
    y_pred = model.predict(X_test)
    assert np.allclose(y_pred, y_test, rtol=1e-2, atol=6.1e-3)
    model.resize_coeff((resize * nx, resize * ny))
    for i in range(4):
        model.coef_
    y_big_pred = model.predict(X_big_test)
    assert np.allclose(y_big_pred, y_big_test, rtol=1e-2, atol=6.1e-2)


def test_setting_kernel():
    from pymks.datasets import make_elastic_FE_strain_delta
    from pymks import MKSLocalizationModel
    from pymks import PrimitiveBasis
    elastic_modulus = (100, 130)
    poissons_ratio = (0.3, 0.3)
    X_delta, y = make_elastic_FE_strain_delta(size=(21, 21),
                                              elastic_modulus=elastic_modulus,
                                              poissons_ratio=poissons_ratio)
    p_basis = PrimitiveBasis(2)
    model = MKSLocalizationModel(basis=p_basis)
    model.fit(X_delta, y)
    coefs = model.coef_
    model.resize_coeff((30, 30))
    model.coef_ = coefs
    assert np.allclose(model.predict(X_delta), y, atol=1e-4)

if __name__ == '__main__':
    test_MKS_elastic_delta()



import pytest
sfepy = pytest.importorskip('sfepy')

import numpy as np
from sfepy.base.goptions import goptions
from sfepy.discrete.fem import Field
try:
    from sfepy.discrete.fem import FEDomain as Domain
except ImportError:
    from sfepy.discrete.fem import Domain
from sfepy.discrete import (FieldVariable, Material, Integral, Function,
                            Equation, Equations, Problem)
from sfepy.terms import Term
from sfepy.discrete.conditions import Conditions, EssentialBC, PeriodicBC
from sfepy.solvers.ls import ScipyDirect
from sfepy.solvers.nls import Newton
import sfepy.discrete.fem.periodic as per
from sfepy.discrete import Functions
from sfepy.mesh.mesh_generators import gen_block_mesh
from sfepy.mechanics.matcoefs import ElasticConstants
from sfepy.base.base import output
from sfepy.discrete.conditions import LinearCombinationBC
from ..fmks.func import deprecate

goptions['verbose'] = False
output.set_output(quiet=True)


class ElasticFESimulation(object):

    """
    Use SfePy to solve a linear strain problem in 2D with a varying
    microstructure on a rectangular grid. The rectangle (cube) is held
    at the negative edge (plane) and displaced by 1 on the positive x
    edge (plane). Periodic boundary conditions are applied to the
    other boundaries.

    The microstructure is of shape (n_samples, n_x, n_y) or (n_samples, n_x,
    n_y, n_z).

    >>> X = np.zeros((1, 3, 3), dtype=int)
    >>> X[0, :, 1] = 1

    >>> sim = ElasticFESimulation(elastic_modulus=(1.0, 10.0),
    ...                           poissons_ratio=(0., 0.))
    >>> sim.run(X)
    >>> y = sim.strain

    y is the strain with components as follows

    >>> exx = y[..., 0]
    >>> eyy = y[..., 1]
    >>> exy = y[..., 2]

    In this example, the strain is only in the x-direction and has a
    uniform value of 1 since the displacement is always 1 and the size
    of the domain is 1.

    >>> assert np.allclose(exx, 1)
    >>> assert np.allclose(eyy, 0)
    >>> assert np.allclose(exy, 0)

    The following example is for a system with contrast. It tests the
    left/right periodic offset and the top/bottom periodicity.

    >>> X = np.array([[[1, 0, 0, 1],
    ...                [0, 1, 1, 1],
    ...                [0, 0, 1, 1],
    ...                [1, 0, 0, 1]]])
    >>> n_samples, N, N = X.shape
    >>> macro_strain = 0.1
    >>> sim = ElasticFESimulation((10.0,1.0), (0.3,0.3), macro_strain=0.1)
    >>> sim.run(X)
    >>> u = sim.displacement[0]

    Check that the offset for the left/right planes is `N *
    macro_strain`.

    >>> assert np.allclose(u[-1,:,0] - u[0,:,0], N * macro_strain)

    Check that the left/right side planes are periodic in y.

    >>> assert np.allclose(u[0,:,1], u[-1,:,1])

    Check that the top/bottom planes are periodic in both x and y.

    >>> assert np.allclose(u[:,0], u[:,-1])

    """
    @deprecate
    def __init__(self, elastic_modulus, poissons_ratio, macro_strain=1.,):
        """Instantiate a ElasticFESimulation.

        Args:
            elastic_modulus (1D array): array of elastic moduli for phases
            poissons_ratio (1D array): array of Possion's ratios for phases
            macro_strain (float, optional): Scalar for macroscopic strain

        """
        self.macro_strain = macro_strain
        self.dx = 1.0
        self.elastic_modulus = elastic_modulus
        self.poissons_ratio = poissons_ratio
        if len(elastic_modulus) != len(poissons_ratio):
            raise RuntimeError(
                'elastic_modulus and poissons_ratio must be the same length')

    def _convert_properties(self, dim):
        """
        Convert from elastic modulus and Poisson's ratio to the Lame
        parameter and shear modulus

        >>> model = ElasticFESimulation(elastic_modulus=(1., 2.),
        ...                             poissons_ratio=(1., 1.))
        >>> result = model._convert_properties(2)
        >>> answer = np.array([[-0.5, 1. / 6.], [-1., 1. / 3.]])
        >>> assert(np.allclose(result, answer))

        Args:
            dim (int): Scalar value for the dimension of the microstructure.

        Returns:
            array with the Lame parameter and the shear modulus for each phase.

        """
        def _convert(E, nu):
            ec = ElasticConstants(young=E, poisson=nu)
            mu = dim / 3. * ec.mu  # pylint: disable=no-member
            lame = ec.lam  # pylint: disable=no-member
            return lame, mu

        return np.array([_convert(E, nu) for E,
                         nu in zip(self.elastic_modulus, self.poissons_ratio)])

    def _get_property_array(self, X):
        """
        Generate property array with elastic_modulus and poissons_ratio for
        each phase.

        Test case for 2D with 3 phases.

        >>> X2D = np.array([[[0, 1, 2, 1],
        ...                  [2, 1, 0, 0],
        ...                  [1, 0, 2, 2]]])
        >>> model2D = ElasticFESimulation(elastic_modulus=(1., 2., 3.),
        ...                               poissons_ratio=(1., 1., 1.))
        >>> lame = lame0, lame1, lame2 = -0.5, -1., -1.5
        >>> mu = mu0, mu1, mu2 = 1. / 6, 1. / 3, 1. / 2
        >>> lm = list(zip(lame, mu))
        >>> X2D_property = np.array([[lm[0], lm[1], lm[2], lm[1]],
        ...                          [lm[2], lm[1], lm[0], lm[0]],
        ...                          [lm[1], lm[0], lm[2], lm[2]]])

        >>> assert(np.allclose(model2D._get_property_array(X2D), X2D_property))

        Test case for 3D with 2 phases.

        >>> model3D = ElasticFESimulation(elastic_modulus=(1., 2.),
        ...                               poissons_ratio=(1., 1.))
        >>> X3D = np.array([[[0, 1],
        ...                  [0, 0]],
        ...                 [[1, 1],
        ...                  [0, 1]]])
        >>> X3D_property = np.array([[[lm[0], lm[1]],
        ...                           [lm[0], lm[0]]],
        ...                          [[lm[1], lm[1]],
        ...                           [lm[0], lm[1]]]])
        >>> assert(np.allclose(model3D._get_property_array(X3D), X3D_property))

        """
        dim = len(X.shape) - 1
        n_phases = len(self.elastic_modulus)
        if not issubclass(X.dtype.type, np.integer):
            raise TypeError("X must be an integer array")
        if np.max(X) >= n_phases or np.min(X) < 0:
            raise RuntimeError(
                "X must be between 0 and {N}.".format(N=n_phases - 1))
        if not (2 <= dim <= 3):
            raise RuntimeError("the shape of X is incorrect")
        return self._convert_properties(dim)[X]

    def run(self, X):
        """
        Run the simulation.

        Args:
          X (ND array): microstructure with shape (n_samples, n_x, ...)
        """
        X_property = self._get_property_array(X)
        strain = []
        displacement = []
        stress = []
        for x in X_property:
            strain_, displacement_, stress_ = self._solve(x)
            strain.append(strain_)
            displacement.append(displacement_)
            stress.append(stress_)
        self.strain = np.array(strain)
        self.displacement = np.array(displacement)
        self.stress = np.array(stress)

    @property
    def response(self):
        return self.strain[..., 0]

    def _get_material(self, property_array, domain):
        """
        Creates an SfePy material from the material property fields for the
        quadrature points.

        Args:
          property_array: array of the properties with shape (n_x, n_y, n_z, 2)

        Returns:
          an SfePy material

        """
        min_xyz = domain.get_mesh_bounding_box()[0]
        dims = domain.get_mesh_bounding_box().shape[1]

        def _material_func_(ts, coors, mode=None, **kwargs):
            if mode == 'qp':
                ijk_out = np.empty_like(coors, dtype=int)
                ijk = np.floor((coors - min_xyz[None]) / self.dx,
                               ijk_out, casting="unsafe")
                ijk_tuple = tuple(ijk.swapaxes(0, 1))
                property_array_qp = property_array[ijk_tuple]
                lam = property_array_qp[..., 0]
                mu = property_array_qp[..., 1]

                from sfepy.mechanics.matcoefs import stiffness_from_lame
                stiffness = stiffness_from_lame(dims, lam=lam, mu=mu)
                lam = np.ascontiguousarray(lam.reshape((lam.shape[0], 1, 1)))
                mu = np.ascontiguousarray(mu.reshape((mu.shape[0], 1, 1)))
                return {'lam': lam, 'mu': mu, 'D': stiffness}
            else:
                return

        material_func = Function('material_func', _material_func_)
        return Material('m', function=material_func)

    def _subdomain_func(self, x=(), y=(), z=(), max_x=None):
        """
        Creates a function to mask subdomains in Sfepy.

        Args:
          x: tuple of lines or points to be masked in the x-plane
          y: tuple of lines or points to be masked in the y-plane
          z: tuple of lines or points to be masked in the z-plane

        Returns:
          array of masked location indices

        """
        eps = 1e-3 * self.dx

        def _func(coords, domain=None):
            flag_x = len(x) == 0
            flag_y = len(y) == 0
            flag_z = len(z) == 0
            for x_ in x:
                flag = (coords[:, 0] < (x_ + eps)) & \
                       (coords[:, 0] > (x_ - eps))
                flag_x = flag_x | flag
            for y_ in y:
                flag = (coords[:, 1] < (y_ + eps)) & \
                       (coords[:, 1] > (y_ - eps))
                flag_y = flag_y | flag
            for z_ in z:
                flag = (coords[:, 2] < (z_ + eps)) & \
                       (coords[:, 2] > (z_ - eps))
                flag_z = flag_z | flag
            flag = flag_x & flag_y & flag_z
            if max_x is not None:
                flag = flag & (coords[:, 0] < (max_x - eps))
            return np.where(flag)[0]

        return _func

    def _get_mesh(self, shape):
        """
        Generate an Sfepy rectangular mesh

        Args:
          shape: proposed shape of domain (vertex shape) (n_x, n_y)

        Returns:
          Sfepy mesh

        """
        center = np.zeros_like(shape)
        return gen_block_mesh(shape, np.array(shape) + 1, center,
                              verbose=False)

    def _get_fixed_displacementsBCs(self, domain):
        """
        Fix the left top and bottom points in x, y and z

        Args:
          domain: an Sfepy domain

        Returns:
          the Sfepy boundary conditions

        """
        min_xyz = domain.get_mesh_bounding_box()[0]
        max_xyz = domain.get_mesh_bounding_box()[1]

        kwargs = {}
        fix_points_dict = {'u.0': 0.0, 'u.1': 0.0}
        if len(min_xyz) == 3:
            kwargs = {'z': (max_xyz[2], min_xyz[2])}
            fix_points_dict['u.2'] = 0.0
        fix_x_points_ = self._subdomain_func(x=(min_xyz[0],),
                                             y=(max_xyz[1], min_xyz[1]),
                                             **kwargs)

        fix_x_points = Function('fix_x_points', fix_x_points_)
        region_fix_points = domain.create_region(
            'region_fix_points',
            'vertices by fix_x_points',
            'vertex',
            functions=Functions([fix_x_points]))
        return EssentialBC('fix_points_BC', region_fix_points, fix_points_dict)

    def _get_shift_displacementsBCs(self, domain):
        """
        Fix the right top and bottom points in x, y and z

        Args:
          domain: an Sfepy domain

        Returns:
          the Sfepy boundary conditions

        """
        min_xyz = domain.get_mesh_bounding_box()[0]
        max_xyz = domain.get_mesh_bounding_box()[1]
        kwargs = {}
        if len(min_xyz) == 3:
            kwargs = {'z': (max_xyz[2], min_xyz[2])}

        displacement = self.macro_strain * (max_xyz[0] - min_xyz[0])
        shift_points_dict = {'u.0': displacement}

        shift_x_points_ = self._subdomain_func(x=(max_xyz[0],),
                                               y=(max_xyz[1], min_xyz[1]),
                                               **kwargs)

        shift_x_points = Function('shift_x_points', shift_x_points_)
        region_shift_points = domain.create_region(
            'region_shift_points',
            'vertices by shift_x_points',
            'vertex',
            functions=Functions([shift_x_points]))
        return EssentialBC('shift_points_BC',
                           region_shift_points, shift_points_dict)

    def _get_displacementBCs(self, domain):
        shift_points_BC = self._get_shift_displacementsBCs(domain)
        fix_points_BC = self._get_fixed_displacementsBCs(domain)
        return Conditions([fix_points_BC, shift_points_BC])

    def _get_linear_combinationBCs(self, domain):
        """
        The right nodes are periodic with the left nodes but also displaced.

        Args:
          domain: an Sfepy domain

        Returns:
          the Sfepy boundary conditions

        """
        min_xyz = domain.get_mesh_bounding_box()[0]
        max_xyz = domain.get_mesh_bounding_box()[1]
        xplus_ = self._subdomain_func(x=(max_xyz[0],))
        xminus_ = self._subdomain_func(x=(min_xyz[0],))

        xplus = Function('xplus', xplus_)
        xminus = Function('xminus', xminus_)
        region_x_plus = domain.create_region('region_x_plus',
                                             'vertices by xplus',
                                             'facet',
                                             functions=Functions([xplus]))
        region_x_minus = domain.create_region('region_x_minus',
                                              'vertices by xminus',
                                              'facet',
                                              functions=Functions([xminus]))
        match_x_plane = Function('match_x_plane', per.match_x_plane)

        def shift_(ts, coors, region):
            return np.ones_like(coors[:, 0]) * \
                self.macro_strain * (max_xyz[0] - min_xyz[0])
        shift = Function('shift', shift_)
        lcbc = LinearCombinationBC(
            'lcbc', [region_x_plus, region_x_minus], {
                'u.0': 'u.0'}, match_x_plane, 'shifted_periodic',
            arguments=(shift,))

        return Conditions([lcbc])

    def _get_periodicBC_X(self, domain, dim):
        dim_dict = {1: ('y', per.match_y_plane),
                    2: ('z', per.match_z_plane)}
        dim_string = dim_dict[dim][0]
        match_plane = dim_dict[dim][1]
        min_, max_ = domain.get_mesh_bounding_box()[:, dim]
        min_x, max_x = domain.get_mesh_bounding_box()[:, 0]
        plus_ = self._subdomain_func(max_x=max_x, **{dim_string: (max_,)})
        minus_ = self._subdomain_func(max_x=max_x, **{dim_string: (min_,)})
        plus_string = dim_string + 'plus'
        minus_string = dim_string + 'minus'
        plus = Function(plus_string, plus_)
        minus = Function(minus_string, minus_)
        region_plus = domain.create_region(
            'region_{0}_plus'.format(dim_string),
            'vertices by {0}'.format(
                plus_string),
            'facet',
            functions=Functions([plus]))
        region_minus = domain.create_region(
            'region_{0}_minus'.format(dim_string),
            'vertices by {0}'.format(
                minus_string),
            'facet',
            functions=Functions([minus]))
        match_plane = Function(
            'match_{0}_plane'.format(dim_string), match_plane)

        bc_dict = {'u.0': 'u.0'}

        bc = PeriodicBC('periodic_{0}'.format(dim_string),
                        [region_plus, region_minus],
                        bc_dict,
                        match='match_{0}_plane'.format(dim_string))
        return bc, match_plane

    def _get_periodicBC_YZ(self, domain, dim):
        dims = domain.get_mesh_bounding_box().shape[1]
        dim_dict = {0: ('x', per.match_x_plane),
                    1: ('y', per.match_y_plane),
                    2: ('z', per.match_z_plane)}
        dim_string = dim_dict[dim][0]
        match_plane = dim_dict[dim][1]
        min_, max_ = domain.get_mesh_bounding_box()[:, dim]
        plus_ = self._subdomain_func(**{dim_string: (max_,)})
        minus_ = self._subdomain_func(**{dim_string: (min_,)})
        plus_string = dim_string + 'plus'
        minus_string = dim_string + 'minus'
        plus = Function(plus_string, plus_)
        minus = Function(minus_string, minus_)
        region_plus = domain.create_region(
            'region_{0}_plus'.format(dim_string),
            'vertices by {0}'.format(
                plus_string),
            'facet',
            functions=Functions([plus]))
        region_minus = domain.create_region(
            'region_{0}_minus'.format(dim_string),
            'vertices by {0}'.format(
                minus_string),
            'facet',
            functions=Functions([minus]))
        match_plane = Function(
            'match_{0}_plane'.format(dim_string), match_plane)

        bc_dict = {'u.1': 'u.1'}
        if dims == 3:
            bc_dict['u.2'] = 'u.2'

        bc = PeriodicBC('periodic_{0}'.format(dim_string),
                        [region_plus, region_minus],
                        bc_dict,
                        match='match_{0}_plane'.format(dim_string))
        return bc, match_plane

    def _solve(self, property_array):
        """
        Solve the Sfepy problem for one sample.

        Args:
          property_array: array of shape (n_x, n_y, 2) where the last
          index is for Lame's parameter and shear modulus,
          respectively.

        Returns:
          the strain field of shape (n_x, n_y, 2) where the last
          index represents the x and y displacements

        """
        shape = property_array.shape[:-1]
        mesh = self._get_mesh(shape)
        domain = Domain('domain', mesh)

        region_all = domain.create_region('region_all', 'all')

        field = Field.from_args('fu', np.float64, 'vector', region_all, # pylint: disable=no-member
                                approx_order=2)

        u = FieldVariable('u', 'unknown', field)
        v = FieldVariable('v', 'test', field, primary_var_name='u')

        m = self._get_material(property_array, domain)

        integral = Integral('i', order=4)

        t1 = Term.new('dw_lin_elastic_iso(m.lam, m.mu, v, u)',
                      integral, region_all, m=m, v=v, u=u)
        eq = Equation('balance_of_forces', t1)
        eqs = Equations([eq])

        epbcs, functions = self._get_periodicBCs(domain)
        ebcs = self._get_displacementBCs(domain)
        lcbcs = self._get_linear_combinationBCs(domain)

        ls = ScipyDirect({})

        pb = Problem('elasticity', equations=eqs, functions=functions)

        pb.time_update(ebcs=ebcs, epbcs=epbcs, lcbcs=lcbcs)

        ev = pb.get_evaluator()
        nls = Newton({}, lin_solver=ls,
                     fun=ev.eval_residual, fun_grad=ev.eval_tangent_matrix)

        try:
            pb.set_solvers_instances(ls, nls)
        except AttributeError:
            pb.set_solver(nls)

        vec = pb.solve()

        u = vec.create_output_dict()['u'].data
        u_reshape = np.reshape(u, (tuple(x + 1 for x in shape) + u.shape[-1:]))

        dims = domain.get_mesh_bounding_box().shape[1]
        strain = np.squeeze(
            pb.evaluate(
                'ev_cauchy_strain.{dim}.region_all(u)'.format(
                    dim=dims),
                mode='el_avg',
                copy_materials=False))
        strain_reshape = np.reshape(strain, (shape + strain.shape[-1:]))

        stress = np.squeeze(
            pb.evaluate(
                'ev_cauchy_stress.{dim}.region_all(m.D, u)'.format(
                    dim=dims),
                mode='el_avg',
                copy_materials=False))
        stress_reshape = np.reshape(stress, (shape + stress.shape[-1:]))

        return strain_reshape, u_reshape, stress_reshape

    def _get_periodicBCs(self, domain):
        dims = domain.get_mesh_bounding_box().shape[1]

        bc_list_YZ, func_list_YZ = list(
            zip(*[self._get_periodicBC_YZ(domain, i) for i in range(0, dims)]))
        bc_list_X, func_list_X = list(
            zip(*[self._get_periodicBC_X(domain, i) for i in range(1, dims)]))
        return Conditions(
            bc_list_YZ + bc_list_X), Functions(func_list_YZ + func_list_X)



from .abstract import _AbstractMicrostructureBasis
from .fftmodule import ifftn, fftn


class _ImagFFTBasis(_AbstractMicrostructureBasis):
    """This class is used to make the bases that create complex valued
    microstructure functions use the standard FFT/iFFT algorithms and selects
    the appropriate fft module depending on whether or not pyfftw is installed.
    """

    def _fftn(self, X):
        """Standard FFT algorithm

        Args:
            X: NDarray (n_samples, N_x, ...)

        Returns:
            Fourier transform of X
        """
        return fftn(X, axes=self._axes, threads=self._n_jobs)

    def _ifftn(self, X):
        """Standard iFFT algorithm

        Args:
            X: NDarray (n_samples, N_x, ...)

        Returns:
            Inverse Fourier transform of X
        """
        return ifftn(X, axes=self._axes, threads=self._n_jobs)



r"""
Legendre basis for microstructure discretization.

Discretize the microstructure function into `n_states` local states
such that:

.. math::
   \frac{1}{\Delta x} \int_s m(h, x) dx =
   \sum_0^{L-1} m[l, s] P_l(h)

where the :math:`P_l` are Legendre polynomials and the local state space
:math:`H` is mapped into the orthogonal domain of the Legendre polynomials

.. math::
   -1 \le  H \le 1

Example.

Here is an example with 4 local states in a microstructure.

>>> data = 2 * da.random.random((1, 3, 3), chunks=(1, 3, 3)) - 1
>>> assert(data.shape == (1, 3, 3))

The when a microstructure is discretized, the different local states are
mapped into local state space, which results in an array of shape
`(n_samples, n_x, n_y, n_states)`, where `n_states=4` in this case.

>>> from toolz import pipe
>>> data_ = pipe(data, discretize(n_state=4, min_=-1, max_=1))
>>> assert(data_.shape == (1, 3, 3, 4))

"""


import numpy as np
import numpy.polynomial.legendre as leg
import dask.array as da
from toolz.curried import pipe
from ..func import curry
from .basis_transformer import BasisTransformer


@curry
def scaled_data(data, domain):
    """
    Scales data to range between -1.0 and 1.0, viz. the domain over which
    legendre polynomials are defined
    """
    return (2.0 * data - domain[0] - domain[1]) / (domain[1] - domain[0])


@curry
def coeff(states):
    """
    Returns a diagonal matrix where in the diagonal terms serve as
    coefficients for individual terms in Legendre polynomial expansion.
    """
    return np.eye(len(states)) * (states + 0.5)


@curry
def leg_data(data, coeff_):
    """
    Computes Legendre expansion for each data point in the
    input data matrix.
    """
    return da.map_blocks(
        lambda x: leg.legval(x, coeff_, tensor=False),
        data,
        chunks=data.chunks[:-1] + (len(coeff_),),
    )


@curry
def discretize(data, n_state=2, min_=0, max_=1, chunks=None):
    """legendre discretization of a microstructure.

    Args:
      data: The microstructure as an `(n_samples, n_x, ...)` shaped
        array where `n_samples` is the number of samples and `n_x` is
        the spatial discretization.
      n_state: the number of local states
      min_: the minimum local state
      max_: the maximum local state
      chunks: the chunks size for the state axis

    Returns:
      Float valued field of of Legendre polynomial coefficients

    >>> data = da.from_array(np.arange(4).reshape(2, 2), chunks=(2, 2))
    >>> out = discretize(data, n_state=3, max_=3, chunks=2)
    >>> print(out.shape)
    (2, 2, 3)
    >>> print(out.chunks)
    ((2,), (2,), (2, 1))
    """
    return pipe(
        data[..., None],
        scaled_data(domain=(min_, max_)),
        leg_data(coeff_=coeff(np.arange(n_state))),
        lambda x: x.rechunk(data.chunks + (chunks or n_state,)),
    )


class LegendreTransformer(BasisTransformer):
    """Legendre transformer for Sklearn pipelines

    >>> from toolz import pipe
    >>> data = da.from_array(np.array([[0, 0.5, 1]]), chunks=(1, 3))
    >>> pipe(
    ...     LegendreTransformer(),
    ...     lambda x: x.fit(None, None),
    ...     lambda x: x.transform(data).compute(),
    ... )
    array([[[ 0.5, -1.5],
            [ 0.5,  0. ],
            [ 0.5,  1.5]]])

    """

    def __init__(self, n_state=2, min_=0.0, max_=1.0, chunks=None):
        """Instantiate a LegendreTransformer

        Args:
          n_state: the number of local states
          min_: the minimum local state
          max_: the maximum local state
          chunks: chunks size for state axis
        """
        super().__init__(
            discretize, n_state=n_state, min_=min_, max_=max_, chunks=chunks
        )



"""MKS Pair Correlation Module
For computing pair correlations from
two point statistics.
"""
import numpy as np
from pymks.fmks.func import (
    fmap,
    sequence,
    star,
    dist_mesh,
    sort_array,
    make_da,
    apply_dict_func,
    curry,
)


def calc_probs(x_data, ind_sort, ind_split):
    """Compute the two point probabilities averages radially

    Args:
       x_data: the two point statistics
       ind_sort: the sorted index of distances from the center
       ind_split: the index groupings

    Returns:
       the calculated probabilites averaged over each radii
       grouping

    """
    return sequence(
        lambda x: x.reshape(x.shape[0], -1)[:, ind_sort],
        lambda x: np.split(x, ind_split[1:], axis=1),
        fmap(lambda x: np.average(x, axis=1)),
        list,
        np.array,
    )(x_data)


def calc_radii(shape):
    """Computer the radii with unique distances from the center

    Args:
      shape: the shape of the domain

    Returns:
      a tuple of (radii, index_split, index_sort) where the
      index_split are groupings of indices at the same distance
      and index_sort are the sorted indices
    """
    flatten = lambda x: x.reshape(-1)
    return sequence(
        dist_mesh,
        flatten,
        sort_array,
        lambda x: np.unique(x[0], return_index=True) + (x[1],),
    )(shape)


def r_inds(cutoff_r, radii):
    """Workout the maximum radii to include

    Args:
      cutoff_r: the cutoff radius (if less than 1 assume it's a
        proportion)
      radii: the radii

    Returns:
      an index of True and False to do mask the proabilities and radii
      for further calculation

    """
    if cutoff_r is None:
        return slice(None)
    if cutoff_r > 1:
        return radii <= cutoff_r
    return radii <= (cutoff_r * radii[-1])


@curry
def calc_cutoff(cutoff_r, probs, radii):
    """Perform the cutoff baseed on cutoff_r

    See `r_inds` above
    """
    index = r_inds(cutoff_r, radii)
    return probs[index], radii[index]


@curry
def interpolate(interpolate_n, probs, radii):
    """Interpoleate the probailites to evenly spaced radii

    Using the max and min radii with `interpolate_n` elements
    interpolate the previously caculated probabilites to new spacing.

    Args:
      interpolate_n: the number of elements
      probs: the probabilites
      radii: the radii

    Returns:
      Tuple of new probabilites and radii

    """
    if interpolate_n is None:
        return probs, radii

    radii_interp = np.linspace(radii.min(), radii.max(), interpolate_n)
    interp = lambda x: np.interp(radii_interp, radii, x)
    return (
        sequence(np.transpose, fmap(interp), list, np.array, np.transpose)(probs),
        radii_interp,
    )


@curry
def np_paircorr(x_data, cutoff_r=None, interpolate_n=None):
    """Numpy only version of `paircorr_from_twopoint`

    """
    return sequence(
        calc_radii,
        lambda x: (calc_probs(x_data, x[2], x[1]), x[0]),
        star(calc_cutoff(cutoff_r)),
        star(interpolate(interpolate_n)),
        lambda x: (np.transpose(x[0]), x[1]),
    )(x_data[0].shape)


@make_da
def paircorr_from_twopoint(x_data, cutoff_r=None, interpolate_n=None):
    r"""Computes the pair correlations from 2-point statistics.

    The pair correlations are the radial average of the 2 point
    stats. The grid spacing is assumed to be one unit. Linear
    interpolation is used if ``interpolate_n`` is specified. If
    another interpolation is desired, don't specify this parameter and
    perform desired interpolation on the output.

    The discretized two point statistics are given by

    .. math::

       f[r \; \vert \; l, l'] = \frac{1}{S} \sum_s m[s, l] m[s + r, l']

    where :math:`f[r \; \vert \; l, l']` is the conditional
    probability of finding the local states :math:`l` and math:`l'` at
    a distance and orientation away from each other defined by the
    vector :math:`r`. `See this paper for more details on the
    notation. <https://doi.org/10.1007/s40192-017-0089-0>`_

    The pair correlation is defined as the conditional probability for
    the case of the magnitude vector, :math:`||r||_2`, defined by :math:`g[d]`.
    :math:`g` is related to :math:`f` via the following
    transformation. Consider the set,
    :math:`I[d] := \{ f[r] \; \vert \; ||r||_2 = d \}`
    then

    .. math::

       g[d] = \frac{1}{ | I[ d ] | }  \sum_{f \in I[ d ]} f

    The :math:`d` are radii from the center pixel of the domain. They
    are automatially calculated if ``interpolate_n`` is ``None``.

    It's assumed that ``x_data`` is a valid set of two point statistics
    calculated from the PyMKS correlations module.

    Args:
      x_data: array of centered 2-point statistics. (n_samples, n_x,
        n_y, ...)
      cutoff_r: the radius cut off. Values less than 1 are assumed to
        be a proportion while values greater than 1 are an exact
        radius cutoff
      interpolate_n: the number of equally spaced radii that the
        probabilities will be interpolated to

    Returns:
      A tuple of the pair correlation array and the radii cutoffs used
      for averaging or interpolation. The pair correlations are shaped
      as ``(n_samples, n_radii)``, whilst the radii are shaped as
      ``(n_radii,)``. ``n_radii`` is equal to ``interpolate_n`` when
      ``interpolate_n`` is specified. The probabilities are chunked on
      the sample axis the same as ``x_data``. The radii is a numpy array.

    Test with only 2 samples of 3x3

    >>> import dask.array as da

    >>> x_data = np.array([
    ...     [
    ...         [0.2, 0.4, 0.3],
    ...         [0.4, 0.5, 0.5],
    ...         [0.2, 0.5, 0.3]
    ...     ],
    ...     [
    ...         [0.1, 0.2, 0.3],
    ...         [0.2, 0.6, 0.4],
    ...         [0.1, 0.4, 0.3]
    ...     ]
    ... ])

    Most basic test

    >>> probs, radii = paircorr_from_twopoint(x_data)
    >>> assert np.allclose(probs,
    ...     [[0.5, 0.45, 0.25],
    ...      [0.6, 0.3, 0.2]])
    >>> assert np.allclose(radii, [0, 1, np.sqrt(2)])

    Test with ``cutoff_r`` greater than 1

    >>> probs, radii = paircorr_from_twopoint(x_data, cutoff_r=1.01)
    >>> assert np.allclose(probs,
    ...     [[0.5, 0.45],
    ...      [0.6, 0.3]])
    >>> assert np.allclose(radii, [0, 1])

    Test with ``cutoff_r`` less than 1

    >>> probs, radii = paircorr_from_twopoint(x_data, cutoff_r=0.99)
    >>> assert np.allclose(probs,
    ...     [[0.5, 0.45],
    ...      [0.6, 0.3]])
    >>> assert np.allclose(radii, [0, 1])

    Test with a linear interpolation

    >>> probs, radii = paircorr_from_twopoint(x_data, interpolate_n=2)
    >>> assert np.allclose(probs,
    ...     [[0.5, 0.25],
    ...      [0.6, 0.2]])
    >>> assert np.allclose(radii, [0, np.sqrt(2)])

    Test with Dask. The chunks along the sample axis are preserved.

    >>> arr = da.from_array(np.random.random((10, 4, 3, 3)), chunks=(2, 4, 3, 3))
    >>> probs, radii = paircorr_from_twopoint(arr)
    >>> probs.shape
    (10, 7)
    >>> probs.chunks
    ((2, 2, 2, 2, 2), (7,))
    >>> assert np.allclose(radii, np.sqrt([0, 1, 2, 3, 4, 5, 6]))

    """
    func = np_paircorr(cutoff_r=cutoff_r, interpolate_n=interpolate_n)
    # Do one sample in serial to calculate the radii and probs shape
    probs, radii = func(x_data[:1].compute())

    return (
        apply_dict_func(
            lambda x: dict(probs=func(x)[0]), x_data, dict(probs=probs.shape)
        )["probs"],
        radii,
    )



"""Test the multiphase microstructure generator

"""

import numpy as np

import pytest

from pymks.fmks.data.multiphase import generate_multiphase, np_generate
from pymks.datasets.microstructure_generator import MicrostructureGenerator


def test_chunking():
    """Test that the generated microstructue is chunked correctly

    """

    data = generate_multiphase(
        shape=(5, 11, 11),
        grain_size=(3, 4),
        volume_fraction=(0.5, 0.5),
        chunks=2,
        seed=10,
    )
    assert data.shape == (5, 11, 11)
    assert data.chunks == ((2, 2, 1), (11,), (11,))


def test_2d():
    """Regression test for microstructure phases
    """
    data = generate_multiphase(
        shape=(1, 4, 4), grain_size=(4, 4), volume_fraction=(0.5, 0.5), seed=10
    )
    assert np.allclose(data, [[[0, 0, 0, 0], [1, 0, 1, 1], [1, 1, 0, 1], [1, 0, 0, 0]]])


def test_1d():
    """Test that 1D works
    """
    data = generate_multiphase(
        shape=(1, 10), grain_size=(4,), volume_fraction=(0.5, 0.5), seed=10
    )
    assert np.allclose(data, [0, 0, 0, 0, 1, 0, 1, 1, 1, 1])


def test_grain_size():
    """
    Test incompatible grain_size and shape
    """
    with pytest.raises(RuntimeError) as excinfo:
        generate_multiphase(
            shape=(1, 10, 10), grain_size=(4,), volume_fraction=(0.5, 0.5), seed=10
        )
    assert str(excinfo.value) == "`shape` should be of length `len(grain_size) + 1`"


def test_volume_fraction():
    """Test incoherent volume_fraction
    """
    with pytest.raises(RuntimeError) as excinfo:
        generate_multiphase(
            shape=(1, 10), grain_size=(2,), volume_fraction=(0.4, 0.4, 0.4), seed=10
        )
    assert str(excinfo.value) == "The terms in the volume fraction list should sum to 1"


def test_3d():
    """Test 3D

    Also tests for bug that occured when chunks and sample size are
    different.

    """
    data = generate_multiphase(
        shape=(5, 5, 5, 5),
        grain_size=(1, 5, 1),
        volume_fraction=(0.3, 0.3, 0.4),
        chunks=2,
        seed=10,
    )
    assert data.chunks == ((2, 2, 1), (5,), (5,), (5,))
    assert np.allclose(
        data[0, 1],
        [
            [0, 1, 0, 2, 2],
            [0, 1, 0, 2, 2],
            [0, 1, 1, 2, 1],
            [0, 1, 2, 2, 0],
            [0, 1, 2, 2, 0],
        ],
    )


def test_versus_old():
    """Test that the fmks microstructure generator gives the same results
    as the previous version.

    """
    n_samples = 3
    size = (4, 4)
    grain_size = (2, 2)
    n_phases = 3
    volume_fraction = (0.5, 0.25, 0.25)
    percent_variance = 0.05

    generator = MicrostructureGenerator(
        n_samples,
        size,
        grain_size=grain_size,
        n_phases=n_phases,
        seed=10,
        volume_fraction=volume_fraction,
        percent_variance=percent_variance,
    )
    x_data_old = generator.generate()

    np.random.seed(10)
    x_data_new = np_generate(
        grain_size,
        volume_fraction,
        percent_variance,
        np.random.random((n_samples,) + size),
    )

    assert np.allclose(x_data_new, x_data_old)



"""MKS localzation module

The MKS localization module fits data using the Materials Knowledge
System in Fourier Space.

Example:

>>> from pymks.fmks.bases.primitive import discretize
>>> from pymks.fmks.func import allclose

>>> disc = discretize(n_state=2)
>>> x_data = lambda: da.from_array(np.linspace(0, 1, 8).reshape((2, 2, 2)),
...                                chunks=(1, 2, 2))
>>> y_data = lambda: x_data().swapaxes(1, 2)
>>> assert pipe(
...     fit(x_data(), y_data(), disc),
...     predict(x_data(), discretize=disc),
...     allclose(y_data())
... )
"""

import dask.array as da
import numpy as np
from scipy.linalg import lstsq
from toolz.curried import pipe
from toolz.curried import map as fmap
from sklearn.base import RegressorMixin, TransformerMixin, BaseEstimator

from .func import (
    curry,
    array_from_tuple,
    rechunk,
    dafftshift,
    dafftn,
    daifftn,
    daifftshift,
    zero_pad,
)


@curry
def lstsq_mks(fx_data, fy_data, redundancy_func, ijk):
    """Do a least square for a single point in the MKS k-space.

    Select a point in k-space using `ijk` tuple and do a least squares
    across samples and local state space to calculate parts of the
    coefficient matrix.

    Args:
      fx_data: microstructure in k-space
      fy_data: response in k-space
      redundancy_func: helps remove redundancies in the coefficient matrix
      ijk: a point in k-space

    Returns:
      a tuple of the coefficient matrix index and vector over local
      state space, which can be used to populate the coefficient
      matrix


    >>> make_data = lambda s: da.from_array(np.arange(np.prod(s),
    ...                                               dtype=float).reshape(s),
    ...                                     chunks=s)

    >>> index, value = lstsq_mks(make_data((2, 1, 1, 3)),
    ...                          make_data((2, 1, 1)),
    ...                          lambda _: (slice(None),),
    ...                          (0, 0))

    >>> print(index)
    (0, 0, slice(None, None, None))
    >>> assert np.allclose(value, [5. / 18., 1. / 9., -1. / 18.])

    """
    fx_data_ = lambda: fx_data[(slice(None),) + ijk + redundancy_func(ijk)]
    fy_data_ = lambda: fy_data[(slice(None),) + ijk]
    return (
        ijk + redundancy_func(ijk),
        lstsq(fx_data_(), fy_data_(), np.finfo(float).eps * 1e4)[0],
    )


def fit_fourier(fx_data, fy_data, redundancy_func):
    """Fit the data in Fourier space.

    Fit the data after it has been discretized and transformed into
    Fourier space.

    Args:
      fx_data: microstructure in k-space
      fy_data: response in k-space
      redundancy_func: helps remove redundancies in the coefficient matrix

    Returns:
      the coefficient matrix (unchunked)

    >>> make_data = lambda s: da.from_array(np.arange(np.prod(s),
    ...                                               dtype=float).reshape(s),
    ...                                     chunks=s)

    >>> matrix = fit_fourier(make_data((5, 4, 4, 3)),
    ...                      make_data((5, 4, 4)),
    ...                      lambda _: (slice(None),))

    >>> print(matrix.shape)
    (4, 4, 3)

    >>> test_matrix = np.resize([5. / 18., 1. / 9., -1. / 18.], (4, 4, 3))
    >>> assert np.allclose(matrix, test_matrix)

    """
    lstsq_mks_ = lstsq_mks(fx_data.compute(), fy_data.compute(), redundancy_func)
    return pipe(
        fmap(lstsq_mks_, np.ndindex(fx_data.shape[1:-1])),
        list,
        array_from_tuple(shape=fx_data.shape[1:], dtype=np.complex),
    )


def faxes(arr):
    """Get the spatiol axes to perform the Fourier transform

    The first and the last axes should not have the Fourier transform
    performed.

    Args:
      arr: the discretized array

    Returns:
      an array starting at 1 to n - 2 where n is the length of the
      shape of arr

    >>> faxes(np.array([1]).reshape((1, 1, 1, 1, 1)))
    array([1, 2, 3])

    """
    return np.arange(arr.ndim - 2) + 1


@curry
def fit_disc(x_data, y_data, redundancy_func):
    """Fit the discretized data.

    Fit the data after the data has already been discretized.

    Args:
      x_data: the discretized mircrostructure field
      y_data: the discretized response field
      redundancy_func: helps remove redundancies in the coefficient matrix

    Returns:
      the chunked coefficient matrix based on the chunking of local
      state space from the discretized mircrostructure field


    >>> make_data = lambda s, c: da.from_array(
    ...     np.arange(np.prod(s),
    ...               dtype=float).reshape(s),
    ...     chunks=c
    ... )

    >>> matrix = fit_disc(make_data((6, 4, 4, 3), (2, 4, 4, 1)),
    ...                   make_data((6, 4, 4), (2, 4, 4)),
    ...                   lambda _: (slice(None),))

    >>> print(matrix.shape)
    (4, 4, 3)

    >>> print(matrix.chunks)
    ((4,), (4,), (1, 1, 1))

    >>> assert np.allclose(matrix.compute()[0, 0, 0], 5. / 18.)

    """
    chunks = lambda x: (None,) * (len(x.shape) - 1) + (x_data.chunks[-1],)
    return pipe(
        [x_data, y_data],
        fmap(dafftn(axes=faxes(x_data))),
        list,
        lambda x: fit_fourier(*x, redundancy_func),
        lambda x: da.from_array(x, chunks=chunks(x)),
    )


@curry
def fit(x_data, y_data, discretize, redundancy_func=lambda _: (slice(None),)):
    """Calculate the MKS influence coefficients.

    Args:
      x_data: the microstructure field
      y_data: the response field
      discretize: a function that returns the discretized data and
        redundancy function


    Returns:
      the influence coefficients

    >>> from pymks.fmks.bases.primitive import discretize

    >>> matrix = fit(da.from_array(np.array([[0], [1]]), chunks=(2, 1)),
    ...              da.from_array(np.array([[2], [1]]), chunks=(2, 1)),
    ...              discretize(n_state=3))
    >>> assert np.allclose(matrix, [[2, 0, 1]])

    """
    return pipe(
        x_data, discretize, fit_disc(y_data=y_data, redundancy_func=redundancy_func)
    )


@curry
def _predict_disc(x_data, coeff):
    return pipe(
        dafftn(x_data, axes=faxes(x_data)),
        lambda x: np.sum(x * coeff[None], axis=-1),
        daifftn(axes=faxes(x_data), s=x_data.shape[1:-1]),
    ).real


@curry
def predict(x_data, coeff, discretize):
    """Predict a response given a microstructure

    Args:
      x_data: the microstructure data
      coeff: the influence coefficients
      discretize: the basis function

    Returns:
      the response
    """
    return _predict_disc(discretize(x_data), coeff)


def _ini_axes(arr):
    return tuple(np.arange(arr.ndim - 1))


@curry
def coeff_to_real(coeff, new_shape=None):
    r"""Convert the coefficients to real space

    Convert the :class:`pymks.LocalizationRegressor` coefficiencts to
    real space. The coefficiencts are calculated in Fourier space, but
    best viewed in real space. If the Fourier coefficients are defined
    as :math:`\beta\left[l, k\right]` then the real space coefficients
    are calculated using,

    .. math::

       \alpha \left[l, r\right] = \frac{1}{N} \sum_{k=0}^{N-1} \beta\left[l, k\right] e^{i \frac{2 \pi}{N} k r} e^{i \pi}

    where :math:`l` is the local state and :math:`r` is the spatial
    index from :math:`0` to :math:`N-1`.  The :math:`e^{i \pi}` term
    is a shift applied to place the 0 coefficient at the center of the
    domain for viewing purposes.

    Args:
      coeff (array): the localization coefficients in Fourier space as a Dask
        array `(n_x, n_y, n_state)`
      new_shape (tuple): shape of the output to either shorten or pad with
        zeros

    Returns:
      the coefficients in real space

    A spike at :math:`k=1` should result in a cosine function on the
    real axis.

    >>> N = 100
    >>> fcoeff = np.zeros((N, 1))
    >>> fcoeff[1] = N
    >>> x = np.linspace(0, 1, N + 1)[:-1]
    >>> assert np.allclose(
    ...     coeff_to_real(da.from_array(fcoeff)).real.compute(),
    ...     np.cos(2 * np.pi * x + np.pi)[:, None]
    ... )

    """
    return pipe(
        coeff,
        daifftn(axes=_ini_axes(coeff), s=new_shape),
        dafftshift(axes=_ini_axes(coeff)),
    )


@curry
def coeff_to_frequency(coeff):
    """Convert the coefficients to frequency space.

    Args:
      coeff: the influence coefficients in real space

    Returns:
      the influence coefficiencts in frequency space

    >>> from .func import rcompose
    >>> f = rcompose(
    ...     lambda x: np.concatenate((x, np.ones_like(x)), axis=-1),
    ...     lambda x: da.from_array(x, chunks=x.shape),
    ...     coeff_to_frequency,
    ...     coeff_to_real,
    ...     lambda x: x.real[..., :1].compute()
    ... )
    >>> assert (lambda x: np.allclose(f(x), x))(np.arange(20).reshape((5, 4, 1)))

    """
    return pipe(
        coeff.copy(), daifftshift(axes=_ini_axes(coeff)), dafftn(axes=_ini_axes(coeff))
    )


@curry
def coeff_resize(coeff, shape):
    """Resize the influence coefficients.

    Resize the influence coefficients by padding with zeros to the
    size determined by shape. Apply to coefficients in frequency space.

    Args:
      coeff: the influence coefficients with size (nx, ny, nz, nstate)
      shape: the new padded shape (NX, NY, NZ)

    Returns:
      the resized influence coefficients

    >>> from .func import ifftshift, fftn
    >>> assert pipe(
    ...     np.arange(20).reshape((5, 4, 1)),
    ...     lambda x: np.concatenate((x, np.ones_like(x)), axis=-1),
    ...     ifftshift(axes=(0, 1)),
    ...     fftn(axes=(0, 1)),
    ...     lambda x: da.from_array(x, chunks=x.shape),
    ...     coeff_resize(shape=(10, 7)),
    ...     coeff_to_real,
    ...     lambda x: np.allclose(x.real[..., 0],
    ...         [[0, 0, 0, 0, 0, 0, 0],
    ...          [0, 0, 0, 0, 0, 0, 0],
    ...          [0, 0, 0, 0, 0, 0, 0],
    ...          [0, 0, 0, 1, 2, 3, 0],
    ...          [0, 0, 4, 5, 6, 7, 0],
    ...          [0, 0, 8, 9,10,11, 0],
    ...          [0, 0,12,13,14,15, 0],
    ...          [0, 0,16,17,18,19, 0],
    ...          [0, 0, 0, 0, 0, 0, 0],
    ...          [0, 0, 0, 0, 0, 0, 0]])
    ... )

    """
    return pipe(
        coeff,
        coeff_to_real,
        zero_pad(
            shape=shape + coeff.shape[-1:],
            chunks=((-1,) * len(shape)) + (coeff.chunks[-1],),
        ),
        coeff_to_frequency,
    )


def reshape(data, shape):
    """Reshape data along all but the first axis

    Args:
        data: the data to reshape
        shape: the shape of the new data (not including the first axis)

    Returns:
        the reshaped data

    >>> data = np.arange(18).reshape((2, 9))
    >>> reshape(data, (None, 3, 3)).shape
    (2, 3, 3)
    """
    return data.reshape(data.shape[0], *shape[1:])


class ReshapeTransformer(BaseEstimator, TransformerMixin):
    """Reshape data ready for the LocalizationRegressor

    Sklearn likes flat image data, but MKS expects shaped data. This
    class transforms the shape of flat data into shaped image data for
    MKS.

    >>> data = np.arange(18).reshape((2, 9))
    >>> ReshapeTransformer((None, 3, 3)).fit(None, None).transform(data).shape
    (2, 3, 3)

    """

    def __init__(self, shape):
        """Instantiate a ReshapeTransformer

        Args:
            shape: the shape of the reshaped data (ignoring the first axis)
        """
        self.shape = shape

    def transform(self, x_data):
        """Transform the X data

        Args:
          x_data: the data to be transformed
        """
        return reshape(x_data, self.shape)

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self


class LocalizationRegressor(BaseEstimator, RegressorMixin):
    """Perform the localization in Sklearn pipelines

    Allows the localization to be part of a Sklearn pipeline

    >>> make_data = lambda s, c: da.from_array(
    ...     np.arange(np.prod(s),
    ...               dtype=float).reshape(s),
    ...     chunks=c
    ... )

    >>> X = make_data((6, 4, 4, 3), (2, 4, 4, 1))
    >>> y = make_data((6, 4, 4), (2, 4, 4))

    >>> y_out = LocalizationRegressor().fit(X, y).predict(X)

    >>> assert np.allclose(y, y_out)

    >>> print(
    ...     pipe(
    ...         LocalizationRegressor(),
    ...         lambda x: x.fit(X, y.reshape(6, 16)).predict(X).shape
    ...     )
    ... )
    (6, 16)

    """

    def __init__(self, redundancy_func=lambda _: (slice(None),)):
        """Instantiate a LocalizationRegressor

        Args:
          redundancy_func: function to remove redundant elements from
            the coefficient matrix

        """
        self.redundancy_func = redundancy_func
        self.coeff = None
        self.y_data_shape = None

    def fit(self, x_data, y_data):
        """Fit the data

        Args:
          x_data: the X data to fit
          y_data: the y data to fit

        Returns:
          the fitted LocalizationRegressor
        """
        self.y_data_shape = y_data.shape
        y_data_reshape = reshape(y_data, x_data.shape[:-1])
        y_data_da = rechunk(x_data.chunks[:-1], y_data_reshape)
        self.coeff = fit_disc(x_data, y_data_da, self.redundancy_func)
        return self

    def predict(self, x_data):
        """Predict the data

        Args:
            x_data: the X data to predict

        Returns:
            The predicted y data
        """
        if len(self.y_data_shape) == len(x_data.shape) - 1:
            new_shape = (1,) + self.coeff.shape[:-1]
        else:
            new_shape = (1, np.prod(self.coeff.shape[:-1]))
        return reshape(_predict_disc(x_data, self.coeff), new_shape)

    def coeff_resize(self, shape):
        """Generate new model with larger coefficients

        Args:
          shape: the shape of the new coefficients

        Returns:
          a new model with larger influence coefficients
        """
        self.coeff = coeff_resize(self.coeff, shape)
        return self
